# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for calculating limiting fluxes.
"""

import warnings

import astropy.units as u
import numpy as np
from astropy.utils.exceptions import AstropyUserWarning
from scipy.ndimage import binary_dilation

from photutils.utils._coords import apply_separation
from photutils.utils._deprecation import (deprecated_getattr,
                                          deprecated_renamed_argument)
from photutils.utils._parameters import (SigmaClipSentinelDefault,
                                         create_default_sigmaclip)
from photutils.utils._progress_bars import add_progress_bar
from photutils.utils._repr import make_repr
from photutils.utils.footprints import circular_footprint

__all__ = ['ImageDepth']

__doctest_requires__ = {('ImageDepth', 'ImageDepth.*'): ['skimage']}

SIGMA_CLIP = SigmaClipSentinelDefault(sigma=3.0, maxiters=10)

# Remove in 4.0
_DEPRECATED_ATTRIBUTES = {
    'nsigma': 'n_sigma',
    'napers': 'n_apertures',
    'niters': 'n_iters',
    'napers_used': 'n_apertures_used',
}


class ImageDepth:
    r"""
    Class to calculate the limiting flux and magnitude of an image.

    Parameters
    ----------
    aper_radius : float
        The radius (in pixels) of the circular apertures used to compute
        the image depth.

    n_sigma : float, optional
        The number of standard deviations at which to compute the image
        depths.

        .. deprecated:: 3.0
            The ``nsigma`` keyword is deprecated. Use ``n_sigma``
            instead.

    mask_pad : float, optional
        An additional padding (in pixels) to apply when dilating the
        input mask.

    n_apertures : int, optional
        The number of circular apertures used to compute the image
        depth.

        .. deprecated:: 3.0
            The ``napers`` keyword is deprecated. Use ``n_apertures``
            instead.

    n_iters : int, optional
        The number of iterations, each with randomly-generated
        apertures, for which the image depth will be calculated.

        .. deprecated:: 3.0
            The ``niters`` keyword is deprecated. Use ``n_iters``
            instead.

    overlap : bool, optional
        Whether to allow the apertures to overlap.

    overlap_maxiters : int, optional
        The maximum number of iterations that will be used when
        attempting to find additional non-overlapping apertures. This
        keyword has no effect unless ``overlap=False``. While increasing
        this number may generate more non-overlapping apertures in
        crowded cases, it will also run slower.

    seed : int, optional
        A seed to initialize the `numpy.random.BitGenerator`. If `None`,
        then fresh, unpredictable entropy will be pulled from the OS.
        Separate function calls with the same ``seed`` will generate the
        same results.

    zeropoint : float, optional
        The zeropoint used to calculate the magnitude limit from the
        flux limit:

        .. math::

            m_{\mathrm{lim}} = -2.5 \log_{10} f_{\mathrm{lim}}
                + \mathrm{zeropoint}

    sigma_clip : `astropy.stats.SigmaClip`, optional
        A `~astropy.stats.SigmaClip` object that defines the sigma
        clipping parameters to use when computing the limiting flux. If
        `None` then no sigma clipping will be performed.

    progress_bar : bool, optional
        Whether to display a progress bar. The progress bar requires
        that the `tqdm <https://tqdm.github.io/>`_ optional dependency
        be installed.

    Attributes
    ----------
    apertures : list of `~photutils.aperture.CircularAperture`
        A list of circular apertures for each iteration.

    napers_used : 1D `~numpy.ndarray`
        .. deprecated:: 3.0
            Use ``n_apertures_used`` instead.

    n_apertures_used : 1D `~numpy.ndarray`
        An array of the number of apertures used for each iteration.

    fluxes : list of `~numpy.ndarray`
        A list of arrays containing the flux measurements for each
        iteration.

    flux_limits : 1D `~numpy.ndarray`
        An array of the flux limits for each iteration.

    mag_limits : 1D `~numpy.ndarray`
        An array of the magnitude limits for each iteration.

    Notes
    -----
    The image depth is calculated by placing random circular apertures
    with the specified radius on blank regions of the image. The number
    of apertures is specified by the ``n_apertures`` keyword. The blank
    regions are calculated from an input mask, which should mask both
    sources in the image and areas without image coverage. The input
    mask will be dilated with a circular footprint with a radius equal
    to the input ``aper_radius`` plus ``mask_pad``. The image border
    is also masked with the same radius.

    The flux limit is calculated as the standard deviation of the
    aperture fluxes times the input ``n_sigma`` significance level. The
    aperture flux values can be sigma clipped prior to computing the
    standard deviation using the ``sigma_clip`` keyword.

    The flux limit is calculated ``n_iters`` times, each with a
    randomly-generated set of circular apertures. The returned flux
    limit is the average of these flux limits.

    The magnitude limit is calculated from flux limit using the input
    ``zeropoint`` keyword as:

    .. math::

        m_{\mathrm{lim}} = -2.5 \log_{10} f_{\mathrm{lim}}
            + \mathrm{zeropoint}

    Examples
    --------
    >>> from astropy.convolution import convolve
    >>> from astropy.visualization import simple_norm
    >>> from photutils.datasets import make_100gaussians_image
    >>> from photutils.segmentation import SourceFinder, make_2dgaussian_kernel
    >>> from photutils.utils import ImageDepth
    >>> bkg = 5.0
    >>> data = make_100gaussians_image() - bkg
    >>> kernel = make_2dgaussian_kernel(3.0, size=5)
    >>> convolved_data = convolve(data, kernel)
    >>> n_pixels = 10
    >>> threshold = 3.2
    >>> finder = SourceFinder(n_pixels=n_pixels, progress_bar=False)
    >>> segment_map = finder(convolved_data, threshold)
    >>> mask = segment_map.make_source_mask()
    >>> radius = 4
    >>> depth = ImageDepth(radius, n_sigma=5.0, n_apertures=500,
    ...                    n_iters=2, mask_pad=5, overlap=False,
    ...                    seed=123, zeropoint=23.9,
    ...                    progress_bar=False)
    >>> limits = depth(data, mask)
    >>> print(np.array(limits))  # doctest: +FLOAT_CMP
    [68.7403149  19.30697121]

    .. plot::
        :include-source:

        # Plot the random apertures for the first iteration

        import matplotlib.pyplot as plt
        from astropy.convolution import convolve
        from astropy.visualization import simple_norm
        from photutils.datasets import make_100gaussians_image
        from photutils.segmentation import SourceFinder, make_2dgaussian_kernel
        from photutils.utils import ImageDepth

        bkg = 5.0
        data = make_100gaussians_image() - bkg
        kernel = make_2dgaussian_kernel(3.0, size=5)
        convolved_data = convolve(data, kernel)
        n_pixels = 10
        threshold = 3.2
        finder = SourceFinder(n_pixels=n_pixels, progress_bar=False)
        segment_map = finder(convolved_data, threshold)
        mask = segment_map.make_source_mask()
        radius = 4
        depth = ImageDepth(radius, n_sigma=5.0, n_apertures=500,
                           n_iters=2, overlap=False, seed=123,
                           progress_bar=False)
        limits = depth(data, mask)

        fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(5, 7))
        norm = simple_norm(data, 'sqrt', percent=99.5)
        ax[0].imshow(data, norm=norm, origin='lower')
        color = 'white'
        depth.apertures[0].plot(ax=ax[0], color=color)
        ax[0].set_title('Data with blank apertures')
        ax[1].imshow(mask, origin='lower')
        depth.apertures[0].plot(ax=ax[1], color=color)
        ax[1].set_title('Mask with blank apertures')

        fig.tight_layout()
    """

    @deprecated_renamed_argument('nsigma', 'n_sigma', '3.0', until='4.0')
    @deprecated_renamed_argument('napers', 'n_apertures', '3.0', until='4.0')
    @deprecated_renamed_argument('niters', 'n_iters', '3.0', until='4.0')
    def __init__(self, aper_radius, *, n_sigma=5.0, mask_pad=0,
                 n_apertures=1000, n_iters=10, overlap=True,
                 overlap_maxiters=100, seed=None, zeropoint=0.0,
                 sigma_clip=SIGMA_CLIP, progress_bar=True):

        if aper_radius <= 0:
            msg = 'aper_radius must be > 0'
            raise ValueError(msg)
        if mask_pad < 0:
            msg = 'mask_pad must be >= 0'
            raise ValueError(msg)

        self.aper_radius = aper_radius
        self.n_sigma = n_sigma
        self.mask_pad = mask_pad
        self.n_apertures = n_apertures
        self.n_iters = n_iters
        self.overlap = overlap
        self.overlap_maxiters = overlap_maxiters
        self.seed = seed
        self.zeropoint = zeropoint
        if sigma_clip is SIGMA_CLIP:
            sigma_clip = create_default_sigmaclip(sigma=SIGMA_CLIP.sigma,
                                                  maxiters=SIGMA_CLIP.maxiters)
        if sigma_clip is not None and not callable(sigma_clip):
            msg = 'sigma_clip must be a callable (e.g., SigmaClip) or None'
            raise TypeError(msg)
        self.sigma_clip = sigma_clip
        self.progress_bar = progress_bar

        self.rng = np.random.default_rng(self.seed)
        self.dilate_radius = int(np.ceil(self.aper_radius + self.mask_pad))
        self.dilate_footprint = circular_footprint(radius=self.dilate_radius)

        self.apertures = []
        self.n_apertures_used = np.array([])
        self.fluxes = []
        self.flux_limits = np.array([])
        self.mag_limits = np.array([])

    def __repr__(self):
        params = ('aper_radius', 'n_sigma', 'mask_pad', 'n_apertures',
                  'n_iters', 'overlap', 'overlap_maxiters', 'seed',
                  'zeropoint', 'sigma_clip', 'progress_bar')
        return make_repr(self, params)

    # Remove in 4.0
    def __getattr__(self, name):
        return deprecated_getattr(self, name,
                                  _DEPRECATED_ATTRIBUTES,
                                  since='3.0', until='4.0')

    def __call__(self, data, mask):
        """
        Calculate the limiting flux and magnitude of an image.

        Parameters
        ----------
        data : 2D `~numpy.ndarray`
            The 2D array, which should be in flux units (not surface
            brightness units).

        mask : 2D bool `~numpy.ndarray`
            A 2D mask array with the same shape as ``data`` where
            a `True` value indicates the corresponding element of
            ``data`` is masked. The input array should mask both sources
            (e.g., from a segmentation image) and regions without image
            coverage. If `None`, then the entire image will be used.

        Returns
        -------
        flux_limit, mag_limit : float
            The flux and magnitude limits. The flux limit is returned in
            the same units as the input ``data``. The magnitude limit is
            calculated from the flux limit and the input ``zeropoint``.
        """
        # Prevent circular import
        from photutils.aperture import CircularAperture

        if mask is None or not np.any(mask):
            all_xycoords = self._make_all_coords_no_mask(data.shape)
        else:
            all_xycoords = self._make_all_coords(mask)

        if len(all_xycoords) == 0:
            msg = ('There are no unmasked pixel values (including the '
                   'masked image borders).')
            raise ValueError(msg)

        n_apertures = self.n_apertures
        if not self.overlap:
            n_apertures2 = 1.5 * self.n_apertures
            n_apertures = int(min(n_apertures2,
                                  0.1 * len(all_xycoords)))

        iter_range = range(self.n_iters)
        if self.progress_bar:
            desc = 'Image Depths'
            iter_range = add_progress_bar(iter_range,
                                          desc=desc)

        flux_limits = []
        apertures = []
        for _ in iter_range:
            if self.overlap:
                xycoords = self._make_coords(all_xycoords, n_apertures)
            else:
                # Cut the number of coords (only need to input ~10x)
                xycoords = self._make_coords(all_xycoords,
                                             n_apertures * 10)
                min_separation = self.aper_radius * 2.0
                xycoords = apply_separation(xycoords, min_separation)
                xycoords = xycoords[0:self.n_apertures]

            apers = CircularAperture(xycoords, r=self.aper_radius)
            apertures.append(apers)
            fluxes, _ = apers.do_photometry(data)
            if self.sigma_clip is not None:
                fluxes = self.sigma_clip(fluxes, masked=False)  # ndarray
            self.fluxes.append(fluxes)
            flux_limits.append(self.n_sigma * np.std(fluxes))

        self.apertures = apertures
        n_apertures_used = np.array([len(apers) for apers in apertures])
        self.n_apertures_used = n_apertures_used
        if np.any(n_apertures_used < self.n_apertures):
            msg = (f'Unable to generate {self.n_apertures} '
                   'non-overlapping apertures in unmasked regions. '
                   'The number of apertures used was less than '
                   f'{self.n_apertures} (see the '
                   '"n_apertures_used" ImageDepth object attribute). '
                   'To fix this, decrease the number of apertures '
                   'and/or aperture size, or increase '
                   '`overlap_maxiters`. Alternatively, you may set '
                   'overlap=True')
            warnings.warn(msg, AstropyUserWarning)

        if isinstance(flux_limits[0], u.Quantity):
            units = True
            self.flux_limits = u.Quantity(flux_limits)
        else:
            units = False
            self.flux_limits = np.array(flux_limits)
        flux_limit = np.mean(self.flux_limits)
        if np.any(self.flux_limits == 0):
            msg = ('One or more flux_limit values was zero. This is '
                   'likely due to constant image values. Check the '
                   'input mask.')
            warnings.warn(msg, AstropyUserWarning)

        # Ignore divide-by-zero RuntimeWarning in log10
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            flux_limits = self.flux_limits
            flux_limit_ = flux_limit
            if units:
                flux_limits = flux_limits.value
                flux_limit_ = flux_limit.value
            self.mag_limits = -2.5 * np.log10(flux_limits) + self.zeropoint
            mag_limit = -2.5 * np.log10(flux_limit_) + self.zeropoint

        return flux_limit, mag_limit

    @staticmethod
    def _find_slice_axis(data, axis):
        """
        Calculate a slice for the minimal bounding box along an axis for
        the `True` values of a 2D boolean array.

        Parameters
        ----------
        data : 2D bool `~numpy.ndarray`
            The boolean array.
        axis : int
            The axis to use (0 or 1).

        Returns
        -------
        slice : slice object
            A slice object for the input axis. If the data values along
            the input axis are all `False`, then the slice object will
            include the entire axis range.
        """
        xx = np.any(data, axis=axis)
        if np.all(~xx):
            idx = 0 if axis else 1
            slc = slice(0, data.shape[idx])
        else:
            x0, x1 = np.where(xx)[0][[0, -1]]
            slc = slice(x0, x1 + 1)

        return slc

    def _find_slices(self, data):
        """
        Calculate a tuple slice for the minimal bounding box for the
        `True` values of a 2D boolean array.

        Parameters
        ----------
        data : 2D bool `~numpy.ndarray`
            The boolean array.

        Returns
        -------
        slices : tuple of slices
            A tuple of slice objects for each axis of the array. If the
            data is all `False`, then the slice tuple will include the
            entire image range.
        """
        xslice = self._find_slice_axis(data, 0)
        yslice = self._find_slice_axis(data, 1)
        return yslice, xslice

    def _mask_border(self, mask):
        """
        Mask pixels around the image border.

        Parameters
        ----------
        mask : 2D bool `~numpy.ndarray`
            Boolean mask array.

        Returns
        -------
        mask : 2D bool `~numpy.ndarray`
            Boolean mask array.
        """
        mask[:self.dilate_radius, :] = True
        mask[-self.dilate_radius:, :] = True
        mask[:, :self.dilate_radius] = True
        mask[:, -self.dilate_radius:] = True

        return mask

    def _dilate_mask(self, mask):
        """
        Dilate the input mask to ensure that apertures do not overlap
        the mask.

        The mask is dilated with a circular footprint with a radius
        equal to the input ``aper_radius`` plus ``mask_pad``.

        Border pixels are also masked with the same radius.

        Parameters
        ----------
        mask : 2D bool `~numpy.ndarray`
            Boolean mask array.

        Returns
        -------
        mask : 2D bool `~numpy.ndarray`
            Dilated boolean mask array.
        """
        mask = np.asarray(mask, dtype=bool).copy()
        if np.any(mask):
            mask = binary_dilation(mask, structure=self.dilate_footprint)

        return self._mask_border(mask)

    def _make_all_coords_no_mask(self, shape):
        """
        Return an array of all possible (x, y) coordinates.

        Border pixels will be excluded.

        Parameters
        ----------
        shape : 2 tuple of int
            The array shape.

        Returns
        -------
        xycoords : 2xN `~numpy.ndarray`
            The (x, y) coordinates.
        """
        ny, nx = shape

        # Remove the image borders
        border = self.dilate_radius
        border2 = 2 * border
        ny -= border2
        nx -= border2

        yi, xi = np.mgrid[0:ny, 0:nx]
        xi = xi.ravel()
        yi = yi.ravel()

        # Shift back to coordinates to the original image
        xi += border
        yi += border

        return np.column_stack((xi, yi))

    def _make_all_coords(self, mask):
        """
        Return an array of all possible unmasked (x, y) coordinates.

        Border pixels will be excluded.

        Parameters
        ----------
        mask : 2D bool `~numpy.ndarray`
            The boolean source mask array.

        Returns
        -------
        xycoords : 2xN `~numpy.ndarray`
            The (x, y) coordinates.
        """
        mask_inv = ~self._dilate_mask(mask)
        mask_slc = self._find_slices(mask_inv)
        yi, xi = np.nonzero(mask_inv[mask_slc])

        # Shift back to coordinates to the original (unsliced) image
        xi += mask_slc[1].start
        yi += mask_slc[0].start

        return np.column_stack((xi, yi))

    def _make_coords(self, xycoords, napers):
        """
        Randomly choose ``napers`` (without replacement) coordinates
        from the input ``xycoords``.

        This function also adds < +/-0.5 pixel random shifts so that the
        coordinates are not all integers.

        Parameters
        ----------
        xycoords : 2xN `~numpy.ndarray`
            The (x, y) coordinates.
        napers : int
            The number of aperture to make.

        Returns
        -------
        xycoords : 2xN `~numpy.ndarray`
            The (x, y) coordinates.
        """
        if napers > xycoords.shape[0]:
            msg = 'Too many apertures for given unmasked area'
            raise ValueError(msg)

        idx = self.rng.choice(xycoords.shape[0], napers, replace=False)
        xycoords = xycoords[idx, :].astype(float)

        shift = self.rng.uniform(-0.5, 0.5, size=xycoords.shape)
        xycoords += shift

        return xycoords



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# cython: language_level=3
"""
Faster evaluation of ellipses from model.py.
"""

import cython
import numpy as np

cimport numpy as cnp

__all__ = ['build_ellipse_model_c']

cnp.import_array()

cdef extern from "math.h":

    double cos(double x)
    double sin(double x)
    double sqrt(double x)


DTYPE = np.float64
ctypedef cnp.float64_t DTYPE_t

cdef inline double get_intens_no_harmonics(double intens0, double phi, double a3, double b3, double a4, double b4):
    return intens0

cdef inline double get_intens_harmonics(double intens0, double phi, double a3, double b3, double a4, double b4):
    return (
        intens0
        + a3 * sin(3.0 * phi)
        + b3 * cos(3.0 * phi)
        + a4 * sin(4.0 * phi)
        + b4 * cos(4.0 * phi)
    )


def build_ellipse_model_c(
    unsigned int n_rows,
    unsigned int n_cols,
    cnp.ndarray[DTYPE_t, ndim=1] finely_spaced_sma,
    cnp.ndarray[DTYPE_t, ndim=1] intens_array,
    cnp.ndarray[DTYPE_t, ndim=1] eps_array,
    cnp.ndarray[DTYPE_t, ndim=1] pa_array,
    cnp.ndarray[DTYPE_t, ndim=1] x0_array,
    cnp.ndarray[DTYPE_t, ndim=1] y0_array,
    cnp.ndarray[DTYPE_t, ndim=1] a3_array = None,
    cnp.ndarray[DTYPE_t, ndim=1] b3_array = None,
    cnp.ndarray[DTYPE_t, ndim=1] a4_array = None,
    cnp.ndarray[DTYPE_t, ndim=1] b4_array = None,
    double phi_min = 0.,
    double phi_max = 2.0*np.pi,
):
    cdef cython.Py_ssize_t len_sma
    len_sma = len(finely_spaced_sma)
    for array in (intens_array, eps_array, pa_array, x0_array, y0_array):
        if len(array) != len_sma:
            raise ValueError(f"All input arrays must be same length={len_sma}")
    harmonic_arrays = (a3_array, b3_array, a4_array, b4_array)
    harmonics_is_none = [array is None for array in harmonic_arrays]

    cdef double a3, b3, a4, b4

    if all(harmonics_is_none):
        intens_func = get_intens_no_harmonics
        a3 = 0
        b3 = 0
        a4 = 0
        b4 = 0
        do_harmonics = False
    else:
        if any(harmonics_is_none):
            raise ValueError("Must supply all harmonic arrays if any is not None")
        for array in harmonic_arrays:
            if len(array) != len_sma:
                raise ValueError(f"All input arrays must be same length={len_sma}")
        intens_func = get_intens_harmonics
        do_harmonics = True

    cdef double phi, fx, fy, x, y
    cdef double one_m_fx, one_m_fy, one_m_fx_t_one_m_fy, one_m_fy_t_fx, one_m_fx_t_fy, fy_t_fx
    cdef double r, sma, q, pa, x0, y0, intens0, intens
    cdef int i, j, i_max, j_max
    cdef cython.Py_ssize_t index
    cdef bool i_ge_zero, i_p1_le_max

    # Define output array
    cdef double[:, :] result = np.zeros([n_rows, n_cols], dtype=DTYPE)
    cdef double[:, :] weight = np.zeros([n_rows, n_cols], dtype=DTYPE)

    i_max = n_cols - 1
    j_max = n_rows - 1

    for index in range(1, len_sma):
        with cython.boundscheck(False):
            sma = finely_spaced_sma[index]
            q = 1.0 - eps_array[index]
            pa = pa_array[index]
            x0 = x0_array[index]
            y0 = y0_array[index]
            intens0 = intens_array[index]
        phi = phi_min
        r = sma
        if do_harmonics:
            with cython.boundscheck(False):
                a3 = a3_array[index]
                b3 = b3_array[index]
                a4 = a4_array[index]
                b4 = b4_array[index]

        while phi <= phi_max:
            # get image coordinates of (r, phi) pixel
            x = r * cos(phi + pa) + x0
            y = r * sin(phi + pa) + y0
            # round down (this is equivalent to int(floor(x)))
            i = int(x) - (x < 0)
            j = int(y) - (y < 0)

            if (-1 <= i <= i_max) and (-1 <= j <= j_max):
                # get fractional deviations relative to target array
                fx = x - float(i)
                fy = y - float(j)

                intens = intens_func(intens0, phi, a3, b3, a4, b4)

                one_m_fx = (1.0 - fx)
                one_m_fy = (1.0 - fy)

                i_ge_zero = i >= 0
                i_p1_le_max = (i + 1) <= i_max

                with cython.boundscheck(False):
                    if j >= 0:
                        if i_ge_zero:
                            weight_pix = one_m_fx * one_m_fy
                            # add up the isophote contribution to the overlapping pixels
                            result[j, i] += intens * weight_pix
                            # add up the fractional area contribution to the
                            # overlapping pixels
                            weight[j, i] += weight_pix
                        if i_p1_le_max:
                            weight_pix = one_m_fy * fx
                            result[j, i + 1] += intens * weight_pix
                            weight[j, i + 1] += weight_pix
                    if (j + 1) <= j_max:
                        if i_ge_zero:
                            weight_pix = one_m_fx * fy
                            result[j + 1, i] += intens * weight_pix
                            weight[j + 1, i] += weight_pix
                        if i_p1_le_max:
                            weight_pix = fy * fx
                            result[j + 1, i + 1] += intens * weight_pix
                            weight[j + 1, i + 1] += weight_pix

            # step towards next pixel on ellipse
            phi = max((phi + 0.75 / r), phi_min)
            r = sma * q / sqrt((q * cos(phi))**2 + sin(phi)**2)
            # max(r, 0.5) could return nan - this is safer
            if not (r >= 0.5):
                r = 0.5

    return np.asarray(result), np.asarray(weight)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
IRAFStarFinder class.
"""

import warnings

import numpy as np
from astropy.utils import lazyproperty
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.detection.core import (_DEPR_DEFAULT, StarFinderBase,
                                      StarFinderCatalogBase,
                                      _handle_deprecated_range,
                                      _StarFinderKernel, _validate_n_brightest)
from photutils.utils._convolution import _filter_data
from photutils.utils._deprecation import (deprecated_positional_kwargs,
                                          deprecated_renamed_argument)
from photutils.utils._quantity_helpers import check_units, isscalar
from photutils.utils._repr import make_repr
from photutils.utils.exceptions import NoDetectionsWarning

__all__ = ['IRAFStarFinder']


class IRAFStarFinder(StarFinderBase):
    """
    Detect stars in an image using IRAF's "starfind" algorithm.

    `IRAFStarFinder` searches images for local density maxima that
    have a peak amplitude greater than ``threshold`` above the local
    background and have a PSF full-width at half-maximum similar to the
    input ``fwhm``. The objects' centroid, roundness (ellipticity), and
    sharpness are calculated using image moments.

    Parameters
    ----------
    threshold : float or 2D `~numpy.ndarray`
        The absolute image value above which to select sources. If
        ``threshold`` is a 2D array, it must have the same shape as the
        input ``data``. If the star finder is run on an image that is a
        `~astropy.units.Quantity` array, then ``threshold`` must have
        the same units.

    fwhm : float
        The full-width half-maximum (FWHM) of the 2D circular Gaussian
        kernel in units of pixels.

    sigma_radius : float, optional
        The truncation radius of the Gaussian kernel in units of sigma
        (standard deviation) (:math:`\\sigma = \\mbox{FWHM} / (2
        \\sqrt{2 \\log(2)})`).

    minsep_fwhm : float, optional
        The separation (in units of ``fwhm``) for detected objects. The
        minimum separation is calculated as ``int((fwhm * minsep_fwhm) +
        0.5)`` and is clipped to a minimum value of 2. Note that large
        values may result in long run times.

        .. deprecated:: 3.0
            Use ``min_separation`` instead.

    sharplo : float, optional
        The lower bound on sharpness for object detection.

        .. deprecated:: 3.0
            Use ``sharpness_range=(lower, upper)`` instead.

    sharphi : float, optional
        The upper bound on sharpness for object detection.

        .. deprecated:: 3.0
            Use ``sharpness_range=(lower, upper)`` instead.

    roundlo : float, optional
        The lower bound on roundness for object detection.

        .. deprecated:: 3.0
            Use ``roundness_range=(lower, upper)`` instead.

    roundhi : float, optional
        The upper bound on roundness for object detection.

        .. deprecated:: 3.0
            Use ``roundness_range=(lower, upper)`` instead.

    exclude_border : bool, optional
        Set to `True` to exclude sources found within half the size of
        the convolution kernel from the image borders. The default is
        `False`, which is the mode used by starfind.

    n_brightest : int, None, optional
        The number of brightest objects to keep after sorting the source
        list by flux. If ``n_brightest`` is set to `None`, all objects
        will be selected.

    peak_max : float, None, optional
        The maximum allowed peak pixel value in an object. Objects with
        peak pixel values greater than ``peak_max`` will be rejected.
        This keyword may be used, for example, to exclude saturated
        sources. If the star finder is run on an image that is a
        `~astropy.units.Quantity` array, then ``peak_max`` must have the
        same units. If ``peak_max`` is set to `None`, then no peak pixel
        value filtering will be performed.

    xycoords : `None` or Nx2 `~numpy.ndarray`, optional
        The (x, y) pixel coordinates of the approximate centroid
        positions of identified sources. If ``xycoords`` are input, the
        algorithm will skip the source-finding step.

    min_separation : `None` or float, optional
        The minimum separation (in pixels) for detected objects. If
        `None` (default) then the minimum separation is calculated as
        ``2.5 * fwhm``. Note that large values may result in long run
        times.

        .. versionchanged:: 3.0
            The default ``min_separation`` changed from ``max(2,
            int(fwhm * 2.5 + 0.5))`` to ``2.5 * fwhm``. To recover the
            previous behavior, set ``min_separation=max(2, int(fwhm *
            2.5 + 0.5))``.

    sharpness_range : tuple of 2 floats or `None`, optional
        The ``(lower, upper)`` inclusive bounds on sharpness for object
        detection. Objects with sharpness outside this range will be
        rejected. If `None`, no sharpness filtering is performed. The
        default is ``(0.5, 2.0)``.

    roundness_range : tuple of 2 floats or `None`, optional
        The ``(lower, upper)`` inclusive bounds on roundness for object
        detection. Objects with roundness outside this range will be
        rejected. If `None`, no roundness filtering is performed. The
        default is ``(0.0, 0.2)``.

    See Also
    --------
    DAOStarFinder

    Notes
    -----
    If the star finder is run on an image that is a
    `~astropy.units.Quantity` array, then ``threshold`` and ``peak_max``
    must have the same units as the image.

    For the convolution step, this routine sets pixels beyond the image
    borders to 0.0. The equivalent parameters in IRAF's starfind are
    ``boundary='constant'`` and ``constant=0.0``.

    IRAF's starfind uses ``hwhmpsf``, ``fradius``, and ``sepmin`` as
    input parameters. The equivalent input values for `IRAFStarFinder`
    are:

    * ``fwhm = hwhmpsf * 2``
    * ``sigma_radius = fradius * sqrt(2.0 * log(2.0))``
    * ``min_separation = max(2, int((fwhm * sepmin) + 0.5))``

    The main differences between `~photutils.detection.DAOStarFinder`
    and `~photutils.detection.IRAFStarFinder` are:

    * `~photutils.detection.IRAFStarFinder` always uses a 2D circular
      Gaussian kernel, while `~photutils.detection.DAOStarFinder` can use
      an elliptical Gaussian kernel.

    * `IRAFStarFinder` internally calculates a "sky" background level
      based on unmasked pixels within the kernel footprint.

    * `~photutils.detection.IRAFStarFinder` calculates the objects'
      centroid, roundness, and sharpness using image moments.
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    @deprecated_renamed_argument('brightest', 'n_brightest', '3.0',
                                 until='4.0')
    @deprecated_renamed_argument('peakmax', 'peak_max', '3.0', until='4.0')
    def __init__(self, threshold, fwhm, sigma_radius=1.5,
                 minsep_fwhm=_DEPR_DEFAULT,
                 sharplo=_DEPR_DEFAULT, sharphi=_DEPR_DEFAULT,
                 roundlo=_DEPR_DEFAULT, roundhi=_DEPR_DEFAULT,
                 exclude_border=False, n_brightest=None, peak_max=None,
                 xycoords=None, min_separation=None, *,
                 sharpness_range=(0.5, 2.0),
                 roundness_range=(0.0, 0.2)):

        # Validate the units, but do not strip them
        inputs = (threshold, peak_max)
        names = ('threshold', 'peak_max')
        check_units(inputs, names)

        if not isscalar(fwhm):
            msg = 'fwhm must be a scalar value'
            raise TypeError(msg)

        sharpness_range = _handle_deprecated_range(
            sharplo, sharphi, sharpness_range,
            'sharp', 'sharpness_range', (0.5, 2.0))
        roundness_range = _handle_deprecated_range(
            roundlo, roundhi, roundness_range,
            'round', 'roundness_range', (0.0, 0.2))

        if sharpness_range is not None:
            if np.ndim(sharpness_range) != 1 or np.size(sharpness_range) != 2:
                msg = ('sharpness_range must be a 2-element (lower, upper) '
                       'tuple or None')
                raise ValueError(msg)
            sharpness_range = tuple(sharpness_range)

        if roundness_range is not None:
            if np.ndim(roundness_range) != 1 or np.size(roundness_range) != 2:
                msg = ('roundness_range must be a 2-element (lower, upper) '
                       'tuple or None')
                raise ValueError(msg)
            roundness_range = tuple(roundness_range)

        # Handle deprecated minsep_fwhm parameter
        if minsep_fwhm is not _DEPR_DEFAULT:
            msg = ("The 'minsep_fwhm' parameter is deprecated "
                   'and will be removed in a future version. Use '
                   "'min_separation' instead.")
            warnings.warn(msg, AstropyDeprecationWarning)
            if minsep_fwhm < 0:
                msg = 'minsep_fwhm must be >= 0'
                raise ValueError(msg)
            if min_separation is None:
                # Use the deprecated minsep_fwhm calculation to set the
                # min_separation
                min_separation = max(2, int((fwhm * minsep_fwhm) + 0.5))

        self.threshold = threshold
        self.fwhm = fwhm
        self.sigma_radius = sigma_radius
        self.sharpness_range = sharpness_range
        self.roundness_range = roundness_range
        self.exclude_border = exclude_border
        self.n_brightest = _validate_n_brightest(n_brightest)
        self.peak_max = peak_max

        if xycoords is not None:
            xycoords = np.asarray(xycoords)
            if xycoords.ndim != 2 or xycoords.shape[1] != 2:
                msg = 'xycoords must be shaped as an Nx2 array'
                raise ValueError(msg)
        self.xycoords = xycoords

        self.kernel = _StarFinderKernel(self.fwhm, ratio=1.0, theta=0.0,
                                        sigma_radius=self.sigma_radius)

        if min_separation is not None:
            if min_separation < 0:
                msg = 'min_separation must be >= 0'
                raise ValueError(msg)
            self.min_separation = min_separation
        else:
            self.min_separation = 2.5 * self.fwhm

    def _repr_str_params(self):
        params = ('threshold', 'fwhm', 'sigma_radius',
                  'sharpness_range', 'roundness_range',
                  'exclude_border', 'n_brightest', 'peak_max', 'xycoords',
                  'min_separation')
        overrides = {}
        if not isscalar(self.threshold):
            overrides['threshold'] = (
                f'<array; shape={np.shape(self.threshold)}>')
        if self.xycoords is not None:
            overrides['xycoords'] = f'<array; shape={self.xycoords.shape}>'
        return params, overrides

    def __repr__(self):
        params, overrides = self._repr_str_params()
        return make_repr(self, params, overrides=overrides)

    def __str__(self):
        params, overrides = self._repr_str_params()
        return make_repr(self, params, overrides=overrides, long=True)

    def _get_raw_catalog(self, data, *, mask=None):
        """
        Get the raw catalog of sources from the input data.

        Parameters
        ----------
        data : 2D `~numpy.ndarray`
            The 2D image array. The image should be
            background-subtracted.

        mask : 2D bool array, optional
            A boolean mask with the same shape as ``data``, where a
            `True` value indicates the corresponding element of ``data``
            is masked. Masked pixels are ignored when searching for
            stars.

        Returns
        -------
        cat : `_IRAFStarFinderCatalog` or `None`
            A catalog of sources found in the input data. `None` is
            returned if no sources are found.
        """
        convolved_data = _filter_data(data, self.kernel.data, mode='constant',
                                      fill_value=0.0,
                                      check_normalization=False)

        if self.xycoords is None:
            xypos = self._find_stars(convolved_data, self.kernel,
                                     self.threshold,
                                     min_separation=self.min_separation,
                                     mask=mask,
                                     exclude_border=self.exclude_border)
        else:
            xypos = self.xycoords

        if xypos is None:
            msg = 'No sources were found.'
            warnings.warn(msg, NoDetectionsWarning)
            return None

        return _IRAFStarFinderCatalog(data,
                                      convolved_data,
                                      xypos,
                                      self.kernel,
                                      sharpness_range=self.sharpness_range,
                                      roundness_range=self.roundness_range,
                                      n_brightest=self.n_brightest,
                                      peak_max=self.peak_max)

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def find_stars(self, data, mask=None):
        """
        Find stars in an astronomical image.

        Parameters
        ----------
        data : 2D array_like
            The 2D image array. The image should be
            background-subtracted.

        mask : 2D bool array, optional
            A boolean mask with the same shape as ``data``, where a
            `True` value indicates the corresponding element of ``data``
            is masked. Masked pixels are ignored when searching for
            stars.

        Returns
        -------
        table : `~astropy.table.QTable` or `None`
            A table of found stars. `None` is returned if no stars are
            found. The table contains the following parameters:

            * ``id``: unique object identification number.
            * ``x_centroid, y_centroid``: object centroid.
            * ``fwhm``: object FWHM.
            * ``sharpness``: object sharpness.
            * ``roundness``: object roundness.
            * ``orientation``: the angle between the ``x`` axis and the
              major axis source measured counter-clockwise in the range
              [0, 360) degrees.
            * ``n_pixels``: the total number of (positive) unmasked
              pixels.
            * ``peak``: the peak, sky-subtracted, pixel value of the object.
            * ``flux``: the object instrumental flux calculated as the
              sum of sky-subtracted data values within the kernel
              footprint.
            * ``mag``: the object instrumental magnitude calculated as
              ``-2.5 * log10(flux)``.
        """
        inputs = (data, self.threshold, self.peak_max)
        names = ('data', 'threshold', 'peak_max')
        check_units(inputs, names)

        cat = self._get_raw_catalog(data, mask=mask)
        if cat is None:
            return None

        # Apply all selection filters
        cat = cat.apply_all_filters()
        if cat is None:
            return None

        # Create the output table
        return cat.to_table()


class _IRAFStarFinderCatalog(StarFinderCatalogBase):
    """
    Class to create a catalog of the properties of each detected star,
    as defined by IRAF's ``starfind`` task.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`
        The 2D image. The image should be background-subtracted.

    convolved_data : 2D `~numpy.ndarray`
        The convolved 2D image. If ``data`` is a
        `~astropy.units.Quantity` array, then ``convolved_data`` must
        have the same units.

    xypos : Nx2 `~numpy.ndarray`
        An Nx2 array of (x, y) pixel coordinates denoting the central
        positions of the stars.

    kernel : `_StarFinderKernel`
        The convolution kernel. This kernel must match the kernel used
        to create the ``convolved_data``.

    sharpness_range : tuple of 2 floats, optional
        The ``(lower, upper)`` inclusive bounds on sharpness for object
        detection. Objects with sharpness outside this range will be
        rejected.

    roundness_range : tuple of 2 floats, optional
        The ``(lower, upper)`` inclusive bounds on roundness for object
        detection. Objects with roundness outside this range will be
        rejected.

    n_brightest : int, None, optional
        The number of brightest objects to keep after sorting the source
        list by flux. If ``n_brightest`` is set to `None`, all objects
        will be selected.

    peak_max : float, None, optional
        The maximum allowed peak pixel value in an object. Objects with
        peak pixel values greater than ``peak_max`` will be rejected.
        This keyword may be used, for example, to exclude saturated
        sources. If the star finder is run on an image that is a
        `~astropy.units.Quantity` array, then ``peak_max`` must have the
        same units. If ``peak_max`` is set to `None`, then no peak pixel
        value filtering will be performed.
    """

    def __init__(self, data, convolved_data, xypos, kernel, *,
                 sharpness_range=(0.2, 1.0), roundness_range=(-1.0, 1.0),
                 n_brightest=None, peak_max=None):

        # Validate the units, but do not strip them
        inputs = (data, convolved_data, peak_max)
        names = ('data', 'convolved_data', 'peak_max')
        check_units(inputs, names)

        super().__init__(data, xypos, kernel,
                         n_brightest=n_brightest,
                         peak_max=peak_max)

        self.convolved_data = convolved_data
        self.sharpness_range = sharpness_range
        self.roundness_range = roundness_range

        self.default_columns = ('id', 'x_centroid', 'y_centroid', 'fwhm',
                                'sharpness', 'roundness', 'orientation',
                                'n_pixels', 'peak', 'flux', 'mag')

    def _get_init_attributes(self):
        """
        Return a tuple of attribute names to copy during slicing.
        """
        return ('data', 'unit', 'convolved_data', 'kernel',
                'sharpness_range', 'roundness_range', 'n_brightest',
                'peak_max', 'cutout_shape', 'default_columns')

    @lazyproperty
    def sky(self):
        """
        Calculate the sky background level.

        The local sky level is roughly estimated using the IRAF starfind
        calculation as the average value in the non-masked regions
        within the kernel footprint.
        """
        skymask = ~self.kernel.mask.astype(bool)  # True=sky, False=obj
        # nsky is always > 0 because the kernel mask never covers the
        # entire footprint (the Gaussian kernel is always truncated
        # within the array, leaving unmasked border pixels).
        nsky = np.count_nonzero(skymask)
        axis = (1, 2)
        sky = np.sum(self.cutout_data_nosub * skymask, axis=axis) / nsky

        if self.unit is not None:
            sky <<= self.unit

        return sky

    @lazyproperty
    def cutout_data_nosub(self):
        """
        The cutout data without sky subtraction or masking.
        """
        return self.make_cutouts(self.data)

    @lazyproperty
    def cutout_data(self):
        """
        The cutout data with sky subtraction and masking applied.
        """
        # This is a freshly computed array, so in-place modification is
        # safe.
        data = ((self.cutout_data_nosub - self.sky[:, np.newaxis, np.newaxis])
                * self.kernel.mask)
        # IRAF starfind discards negative pixels
        data[data < 0] = 0.0
        return data

    @lazyproperty
    def n_pixels(self):
        """
        The total number of (positive) unmasked pixels in the cutout
        data.
        """
        return np.count_nonzero(self.cutout_data, axis=(1, 2))

    @lazyproperty
    def cutout_xorigin(self):
        """
        The x pixel coordinate of the cutout origin.
        """
        return np.transpose(self.xypos)[0] - self.kernel.x_radius

    @lazyproperty
    def cutout_yorigin(self):
        """
        The y pixel coordinate of the cutout origin.
        """
        return np.transpose(self.xypos)[1] - self.kernel.y_radius

    @lazyproperty
    def x_centroid(self):
        """
        The x pixel coordinate of the object centroid.
        """
        return self.cutout_x_centroid + self.cutout_xorigin

    @lazyproperty
    def y_centroid(self):
        """
        The y pixel coordinate of the object centroid.
        """
        return self.cutout_y_centroid + self.cutout_yorigin

    @lazyproperty
    def sharpness(self):
        """
        The sharpness of the object.
        """
        return self.fwhm / self.kernel.fwhm

    def apply_filters(self):
        """
        Filter the catalog.
        """
        attrs = ('x_centroid', 'y_centroid', 'sharpness', 'roundness',
                 'orientation', 'sky', 'peak', 'flux')
        initial_mask = np.count_nonzero(self.cutout_data, axis=(1, 2)) > 1
        newcat = self._filter_finite(attrs, initial_mask=initial_mask)
        if newcat is None:
            return None

        bounds = [
            ('sharpness', self.sharpness_range),
            ('roundness', self.roundness_range),
        ]
        return newcat._filter_bounds(bounds)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Circular and circular-annulus apertures in both pixel and sky
coordinates.
"""

import math

import astropy.units as u
import numpy as np
from astropy.coordinates import Angle
from astropy.utils import lazyproperty

from photutils.aperture.attributes import (PixelPositions, PositiveScalar,
                                           PositiveScalarAngle,
                                           SkyCoordPositions)
from photutils.aperture.core import PixelAperture, SkyAperture
from photutils.aperture.mask import ApertureMask
from photutils.geometry import circular_overlap_grid
from photutils.utils._deprecation import deprecated
from photutils.utils._wcs_helpers import (pixel_to_sky_mean_scale,
                                          sky_to_pixel_mean_scale)

__all__ = [
    'CircularAnnulus',
    'CircularAperture',
    'CircularMaskMixin',
    'SkyCircularAnnulus',
    'SkyCircularAperture',
]


@deprecated('3.0')
class CircularMaskMixin:  # pragma: no cover
    """
    Mixin class to create masks for circular and circular-annulus
    aperture objects.

    .. deprecated:: 3.0
    """

    def to_mask(self, method='exact', subpixels=5):
        """
        Return a mask for the aperture.

        Parameters
        ----------
        method : {'exact', 'center', 'subpixel'}, optional
            The method used to determine the overlap of the aperture
            on the pixel grid. Not all options are available for all
            aperture types. Note that the more precise methods are
            generally slower. The following methods are available:

            * ``'exact'`` (default):
              The exact fractional overlap of the aperture and each
              pixel is calculated. The aperture weights will contain
              values between 0 and 1.

            * ``'center'``:
              A pixel is considered to be entirely in or out of the
              aperture depending on whether its center is in or out of
              the aperture. The aperture weights will contain values
              only of 0 (out) and 1 (in).

            * ``'subpixel'``:
              A pixel is divided into subpixels (see the ``subpixels``
              keyword), each of which are considered to be entirely in
              or out of the aperture depending on whether its center is
              in or out of the aperture. If ``subpixels=1``, this method
              is equivalent to ``'center'``. The aperture weights will
              contain values between 0 and 1.

        subpixels : int, optional
            For the ``'subpixel'`` method, resample pixels by this
            factor in each dimension. That is, each pixel is divided
            into ``subpixels**2`` subpixels. This keyword is ignored
            unless ``method='subpixel'``.

        Returns
        -------
        mask : `~photutils.aperture.ApertureMask` or list of \
                `~photutils.aperture.ApertureMask`
            A mask for the aperture. If the aperture is scalar then
            a single `~photutils.aperture.ApertureMask` is returned,
            otherwise a list of `~photutils.aperture.ApertureMask` is
            returned.
        """
        use_exact, subpixels = self._translate_mask_method(method, subpixels)

        if hasattr(self, 'r'):
            radius = self.r
        elif hasattr(self, 'r_out'):  # annulus
            radius = self.r_out
        else:
            msg = 'Cannot determine the aperture radius'
            raise ValueError(msg)

        masks = []
        for bbox, edges in zip(self._bbox, self._centered_edges, strict=True):
            ny, nx = bbox.shape
            mask = circular_overlap_grid(edges[0], edges[1], edges[2],
                                         edges[3], nx, ny, radius, use_exact,
                                         subpixels)

            # Subtract the inner circle for an annulus
            if hasattr(self, 'r_in'):
                mask -= circular_overlap_grid(edges[0], edges[1], edges[2],
                                              edges[3], nx, ny, self.r_in,
                                              use_exact, subpixels)

            masks.append(ApertureMask(mask, bbox))

        if self.isscalar:
            return masks[0]

        return masks


class CircularAperture(PixelAperture):
    """
    A circular aperture defined in pixel coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : array_like
        The pixel coordinates of the aperture center(s) in one of the
        following formats:

        * single ``(x, y)`` pair as a tuple, list, or `~numpy.ndarray`
        * tuple, list, or `~numpy.ndarray` of ``(x, y)`` pairs

    r : float
        The radius of the circle in pixels.

    Raises
    ------
    ValueError : `ValueError`
        If the input radius, ``r``, is negative.

    Examples
    --------
    >>> from photutils.aperture import CircularAperture
    >>> aper = CircularAperture([10.0, 20.0], 3.0)
    >>> aper = CircularAperture((10.0, 20.0), 3.0)

    >>> pos1 = (10.0, 20.0)  # (x, y)
    >>> pos2 = (30.0, 40.0)
    >>> pos3 = (50.0, 60.0)
    >>> aper = CircularAperture([pos1, pos2, pos3], 3.0)
    >>> aper = CircularAperture((pos1, pos2, pos3), 3.0)
    """

    _params = ('positions', 'r')
    positions = PixelPositions('The center pixel position(s).')
    r = PositiveScalar('The radius in pixels.')

    def __init__(self, positions, r):
        self.positions = positions
        self.r = r

    @lazyproperty
    def _xy_extents(self):
        return self.r, self.r

    @lazyproperty
    def area(self):
        """
        The exact geometric area of the aperture shape.
        """
        return math.pi * self.r**2

    def _to_patch(self, *, origin=(0, 0), **kwargs):
        """
        Return a `~matplotlib.patches.Patch` for the aperture.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` position of the origin of the displayed
            image.

        **kwargs : dict, optional
            Any keyword arguments accepted by
            `matplotlib.patches.Patch`.

        Returns
        -------
        patch : `~matplotlib.patches.Patch` or list of \
                `~matplotlib.patches.Patch`
            A patch for the aperture. If the aperture is scalar then a
            single `~matplotlib.patches.Patch` is returned, otherwise a
            list of `~matplotlib.patches.Patch` is returned.
        """
        import matplotlib.patches as mpatches

        xy_positions, patch_kwargs = self._define_patch_params(origin=origin,
                                                               **kwargs)

        patches = [mpatches.Circle(xy_position, self.r, **patch_kwargs)
                   for xy_position in xy_positions]

        if self.isscalar:
            return patches[0]

        return patches

    def _compute_overlap(self, edges, nx, ny, use_exact, subpixels):
        """
        Compute the overlap of the aperture on the pixel grid.

        Parameters
        ----------
        edges : list of 4 1D `~numpy.ndarray`
            The edges of the pixel grid in the form of
            ``[x_edges, y_edges, x_centers, y_centers]``.

        nx, ny : int
            The number of pixels in the x and y directions.

        use_exact : bool
            Whether to use the exact method for calculating the overlap.

        subpixels : int
            The number of subpixels to use in each dimension for the
            subpixel method.

        Returns
        -------
        overlap : 2D `~numpy.ndarray`
            The overlap of the aperture on the pixel grid. The values
            will be between 0 and 1, where 0 means no overlap and 1
            means full overlap.
        """
        return circular_overlap_grid(edges[0], edges[1], edges[2],
                                     edges[3], nx, ny, self.r,
                                     use_exact, subpixels)

    def to_sky(self, wcs):
        """
        Convert the aperture to a `SkyCircularAperture` object defined
        in celestial coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `SkyCircularAperture` object
            A `SkyCircularAperture` object.

        Notes
        -----
        The aperture shape parameters are converted using the local
        WCS pixel scale evaluated at the first aperture position.
        Because aperture objects require scalar shape parameters, only
        a single reference position is used for the conversion. For
        apertures with multiple positions used with a WCS that has
        spatially-varying distortions, this may produce inaccurate
        results for positions far from the first position.
        """
        xpos, ypos = np.transpose(self.positions)
        positions = wcs.pixel_to_world(xpos, ypos)

        first_pos = np.atleast_2d(self.positions)[0]
        _, mean_scale = pixel_to_sky_mean_scale(
            (float(first_pos[0]), float(first_pos[1])), wcs)

        r = Angle(self.r * mean_scale, 'arcsec')
        return SkyCircularAperture(positions=positions, r=r)


class CircularAnnulus(PixelAperture):
    """
    A circular annulus aperture defined in pixel coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : array_like
        The pixel coordinates of the aperture center(s) in one of the
        following formats:

        * single ``(x, y)`` pair as a tuple, list, or `~numpy.ndarray`
        * tuple, list, or `~numpy.ndarray` of ``(x, y)`` pairs

    r_in : float
        The inner radius of the circular annulus in pixels.

    r_out : float
        The outer radius of the circular annulus in pixels.

    Raises
    ------
    ValueError : `ValueError`
        If inner radius (``r_in``) is greater than outer radius
        (``r_out``).

    ValueError : `ValueError`
        If inner radius (``r_in``) is negative.

    Examples
    --------
    >>> from photutils.aperture import CircularAnnulus
    >>> aper = CircularAnnulus([10.0, 20.0], 3.0, 5.0)
    >>> aper = CircularAnnulus((10.0, 20.0), 3.0, 5.0)

    >>> pos1 = (10.0, 20.0)  # (x, y)
    >>> pos2 = (30.0, 40.0)
    >>> pos3 = (50.0, 60.0)
    >>> aper = CircularAnnulus([pos1, pos2, pos3], 3.0, 5.0)
    >>> aper = CircularAnnulus((pos1, pos2, pos3), 3.0, 5.0)
    """

    _params = ('positions', 'r_in', 'r_out')
    positions = PixelPositions('The center pixel position(s).')
    r_in = PositiveScalar('The inner radius in pixels.')
    r_out = PositiveScalar('The outer radius in pixels.')

    def __init__(self, positions, r_in, r_out):
        if not r_out > r_in:
            msg = "'r_out' must be greater than 'r_in'"
            raise ValueError(msg)

        self.positions = positions
        self.r_in = r_in
        self.r_out = r_out

    @lazyproperty
    def _xy_extents(self):
        return self.r_out, self.r_out

    @lazyproperty
    def area(self):
        """
        The exact geometric area of the aperture shape.
        """
        return math.pi * (self.r_out**2 - self.r_in**2)

    def _to_patch(self, *, origin=(0, 0), **kwargs):
        """
        Return a `~matplotlib.patches.Patch` for the aperture.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` position of the origin of the displayed
            image.

        **kwargs : dict, optional
            Any keyword arguments accepted by
            `matplotlib.patches.Patch`.

        Returns
        -------
        patch : `~matplotlib.patches.Patch` or list of \
                `~matplotlib.patches.Patch`
            A patch for the aperture. If the aperture is scalar then a
            single `~matplotlib.patches.Patch` is returned, otherwise a
            list of `~matplotlib.patches.Patch` is returned.
        """
        import matplotlib.patches as mpatches

        xy_positions, patch_kwargs = self._define_patch_params(origin=origin,
                                                               **kwargs)

        patches = []
        for xy_position in xy_positions:
            patch_inner = mpatches.Circle(xy_position, self.r_in)
            patch_outer = mpatches.Circle(xy_position, self.r_out)
            path = self._make_annulus_path(patch_inner, patch_outer)
            patches.append(mpatches.PathPatch(path, **patch_kwargs))

        if self.isscalar:
            return patches[0]

        return patches

    def _compute_overlap(self, edges, nx, ny, use_exact, subpixels):
        """
        Compute the overlap of the aperture on the pixel grid.

        Parameters
        ----------
        edges : list of 4 1D `~numpy.ndarray`
            The edges of the pixel grid in the form of
            ``[x_edges, y_edges, x_centers, y_centers]``.

        nx, ny : int
            The number of pixels in the x and y directions.

        use_exact : bool
            Whether to use the exact method for calculating the overlap.

        subpixels : int
            The number of subpixels to use in each dimension for the
            subpixel method.

        Returns
        -------
        overlap : 2D `~numpy.ndarray`
            The overlap of the aperture on the pixel grid. The values
            will be between 0 and 1, where 0 means no overlap and 1
            means full overlap.
        """
        overlap = circular_overlap_grid(edges[0], edges[1], edges[2],
                                        edges[3], nx, ny, self.r_out,
                                        use_exact, subpixels)
        overlap -= circular_overlap_grid(edges[0], edges[1], edges[2],
                                         edges[3], nx, ny, self.r_in,
                                         use_exact, subpixels)
        return overlap

    def to_sky(self, wcs):
        """
        Convert the aperture to a `SkyCircularAnnulus` object defined in
        celestial coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `SkyCircularAnnulus` object
            A `SkyCircularAnnulus` object.

        Notes
        -----
        The aperture shape parameters are converted using the local
        WCS pixel scale evaluated at the first aperture position.
        Because aperture objects require scalar shape parameters, only
        a single reference position is used for the conversion. For
        apertures with multiple positions used with a WCS that has
        spatially-varying distortions, this may produce inaccurate
        results for positions far from the first position.
        """
        xpos, ypos = np.transpose(self.positions)
        positions = wcs.pixel_to_world(xpos, ypos)

        first_pos = np.atleast_2d(self.positions)[0]
        _, mean_scale = pixel_to_sky_mean_scale(
            (float(first_pos[0]), float(first_pos[1])), wcs)

        r_in = Angle(self.r_in * mean_scale, 'arcsec')
        r_out = Angle(self.r_out * mean_scale, 'arcsec')
        return SkyCircularAnnulus(positions=positions, r_in=r_in, r_out=r_out)


class SkyCircularAperture(SkyAperture):
    """
    A circular aperture defined in sky coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : `~astropy.coordinates.SkyCoord`
        The celestial coordinates of the aperture center(s). This can be
        either scalar coordinates or an array of coordinates.

    r : scalar `~astropy.units.Quantity`
        The radius of the circle in angular units.

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u
    >>> from photutils.aperture import SkyCircularAperture
    >>> positions = SkyCoord(ra=[10.0, 20.0], dec=[30.0, 40.0], unit='deg')
    >>> aper = SkyCircularAperture(positions, 0.5*u.arcsec)
    """

    _params = ('positions', 'r')
    positions = SkyCoordPositions('The center position(s) in sky coordinates.')
    r = PositiveScalarAngle('The radius in angular units.')

    def __init__(self, positions, r):
        self.positions = positions
        self.r = r

    def to_pixel(self, wcs):
        """
        Convert the aperture to a `CircularAperture` object defined in
        pixel coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `CircularAperture` object
            A `CircularAperture` object.

        Notes
        -----
        The aperture shape parameters are converted using the local
        WCS pixel scale evaluated at the first aperture position.
        Because aperture objects require scalar shape parameters, only
        a single reference position is used for the conversion. For
        apertures with multiple positions used with a WCS that has
        spatially-varying distortions, this may produce inaccurate
        results for positions far from the first position.
        """
        xpos, ypos = wcs.world_to_pixel(self.positions)
        positions = np.transpose((xpos, ypos))

        skypos = self.positions if self.isscalar else self.positions[0]
        _, mean_scale = sky_to_pixel_mean_scale(skypos, wcs)

        r = self.r.to(u.arcsec).value * mean_scale
        return CircularAperture(positions=positions, r=r)


class SkyCircularAnnulus(SkyAperture):
    """
    A circular annulus aperture defined in sky coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : `~astropy.coordinates.SkyCoord`
        The celestial coordinates of the aperture center(s). This can be
        either scalar coordinates or an array of coordinates.

    r_in : scalar `~astropy.units.Quantity`
        The inner radius of the circular annulus in angular units.

    r_out : scalar `~astropy.units.Quantity`
        The outer radius of the circular annulus in angular units.

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u
    >>> from photutils.aperture import SkyCircularAnnulus
    >>> positions = SkyCoord(ra=[10.0, 20.0], dec=[30.0, 40.0], unit='deg')
    >>> aper = SkyCircularAnnulus(positions, 0.5*u.arcsec, 1.0*u.arcsec)
    """

    _params = ('positions', 'r_in', 'r_out')
    positions = SkyCoordPositions('The center position(s) in sky coordinates.')
    r_in = PositiveScalarAngle('The inner radius in angular units.')
    r_out = PositiveScalarAngle('The outer radius in angular units.')

    def __init__(self, positions, r_in, r_out):
        if not r_out > r_in:
            msg = "'r_out' must be greater than 'r_in'"
            raise ValueError(msg)

        self.positions = positions
        self.r_in = r_in
        self.r_out = r_out

    def to_pixel(self, wcs):
        """
        Convert the aperture to a `CircularAnnulus` object defined in
        pixel coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `CircularAnnulus` object
            A `CircularAnnulus` object.

        Notes
        -----
        The aperture shape parameters are converted using the local
        WCS pixel scale evaluated at the first aperture position.
        Because aperture objects require scalar shape parameters, only
        a single reference position is used for the conversion. For
        apertures with multiple positions used with a WCS that has
        spatially-varying distortions, this may produce inaccurate
        results for positions far from the first position.
        """
        xpos, ypos = wcs.world_to_pixel(self.positions)
        positions = np.transpose((xpos, ypos))

        skypos = self.positions if self.isscalar else self.positions[0]
        _, mean_scale = sky_to_pixel_mean_scale(skypos, wcs)

        r_in = self.r_in.to(u.arcsec).value * mean_scale
        r_out = self.r_out.to(u.arcsec).value * mean_scale
        return CircularAnnulus(positions=positions, r_in=r_in, r_out=r_out)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for making tables of model parameters or making models from a
table of model parameters.
"""

import numpy as np
from astropy.table import QTable

from photutils.utils._coords import make_random_xycoords
from photutils.utils._deprecation import deprecated_positional_kwargs
from photutils.utils._misc import _get_meta
from photutils.utils._parameters import as_pair

__all__ = ['make_model_params', 'make_random_models_table',
           'params_table_to_models']


def make_model_params(shape, n_sources, *, x_name='x_0', y_name='y_0',
                      min_separation=1, border_size=(0, 0), seed=0, **kwargs):
    """
    Make a table of randomly generated model positions and additional
    parameters for simulated sources.

    By default, this function computes only a table of x_0 and y_0
    values. Additional parameters can be specified as keyword arguments
    with their lower and upper bounds as 2-tuples. The parameter values
    will be uniformly distributed between the lower and upper bounds,
    inclusively.

    Parameters
    ----------
    shape : 2-tuple of int
        The shape of the output image.

    n_sources : int
        The number of sources to generate. If ``min_separation`` is too
        large, the number of requested sources may not fit within the
        given ``shape`` and therefore the number of sources generated
        may be less than ``n_sources``.

    x_name : str, optional
        The name of the ``model`` parameter that corresponds to the x
        position of the sources. This will be the column name in the
        output table.

    y_name : str, optional
        The name of the ``model`` parameter that corresponds to the y
        position of the sources. This will be the column name in the
        output table.

    min_separation : float, optional
        The minimum separation between the centers of two sources. Note
        that if the minimum separation is too large, the number of
        sources generated may be less than ``n_sources``.

    border_size : tuple of 2 int or int, optional
        The (ny, nx) size of the border around the image where no
        sources will be generated (i.e., the source center will not be
        located within the border). If a single integer is provided, it
        will be used for both dimensions.

    seed : int, optional
        A seed to initialize the `numpy.random.BitGenerator`. If `None`,
        then fresh, unpredictable entropy will be pulled from the OS.

    **kwargs
        Keyword arguments are accepted for additional model parameters.
        The values should be 2-tuples of the lower and upper bounds for
        the parameter range. The parameter values will be uniformly
        distributed between the lower and upper bounds, inclusively.

    Returns
    -------
    table : `~astropy.table.QTable`
        A table containing the model parameters of the generated
        sources. The table will also contain an ``'id'`` column with
        unique source IDs.

    Examples
    --------
    >>> from photutils.datasets import make_model_params
    >>> params = make_model_params((100, 100), 5, flux=(100, 500),
    ...                            min_separation=3, border_size=10, seed=0)
    >>> for col in params.colnames:
    ...     params[col].info.format = '%.8g'  # for consistent table output
    >>> print(params)
     id    x_0       y_0       flux
    --- --------- --------- ---------
      1 60.956935 72.967865 291.99517
      2 31.582937 29.149555 192.94917
      3 13.277882 80.118738 420.75223
      4 11.322211 14.685443 469.41206
      5 75.061619 36.889365 206.45211

    >>> params = make_model_params((100, 100), 5, flux=(100, 500),
    ...                            x_name='x_mean', y_name='y_mean',
    ...                            min_separation=3, border_size=10, seed=0)
    >>> for col in params.colnames:
    ...     params[col].info.format = '%.8g'  # for consistent table output
    >>> print(params)
     id   x_mean    y_mean     flux
    --- --------- --------- ---------
      1 60.956935 72.967865 291.99517
      2 31.582937 29.149555 192.94917
      3 13.277882 80.118738 420.75223
      4 11.322211 14.685443 469.41206
      5 75.061619 36.889365 206.45211

    >>> params = make_model_params((100, 100), 5, flux=(100, 500),
    ...                            sigma=(1, 2), alpha=(0, 1),
    ...                            min_separation=3, border_size=10, seed=0)
    >>> for col in params.colnames:
    ...     params[col].info.format = '%.5g'  # for consistent table output
    >>> print(params)
     id  x_0    y_0    flux  sigma   alpha
    --- ------ ------ ------ ------ --------
      1 60.957 72.968    292 1.5389  0.61437
      2 31.583  29.15 192.95 1.4428 0.028365
      3 13.278 80.119 420.75  1.931  0.71922
      4 11.322 14.685 469.41 1.0405 0.015992
      5 75.062 36.889 206.45  1.732  0.75795
    """
    shape = as_pair('shape', shape, lower_bound=(0, 0))
    border_size = as_pair('border_size', border_size, lower_bound=(0, 1))

    xrange = (border_size[1], shape[1] - border_size[1])
    yrange = (border_size[0], shape[0] - border_size[0])

    if xrange[0] >= xrange[1] or yrange[0] >= yrange[1]:
        msg = 'border_size is too large for the given shape'
        raise ValueError(msg)

    rng = np.random.default_rng(seed)
    xycoords = make_random_xycoords(n_sources, xrange, yrange,
                                    min_separation=min_separation,
                                    seed=rng)
    x, y = np.transpose(xycoords)

    model_params = QTable()
    model_params['id'] = np.arange(len(x)) + 1
    model_params[x_name] = x
    model_params[y_name] = y

    for param, prange in kwargs.items():
        if len(prange) != 2:
            msg = f'{param} must be a 2-tuple'
            raise ValueError(msg)
        vals = rng.uniform(*prange, len(model_params))
        model_params[param] = vals

    return model_params


@deprecated_positional_kwargs(since='3.0', until='4.0')
def make_random_models_table(n_sources, param_ranges, seed=None):
    """
    Make a `~astropy.table.QTable` containing randomly generated
    parameters for an Astropy model to simulate a set of sources.

    Each row of the table corresponds to a source whose parameters are
    defined by the column names. The parameters are drawn from a uniform
    distribution over the specified input ranges, inclusively.

    The output table can be input into :func:`make_model_image` to
    create an image containing the model sources.

    Parameters
    ----------
    n_sources : int
        The number of random model sources to generate.

    param_ranges : dict
        The lower and upper boundaries for each of the model parameters
        as a dictionary mapping the parameter name to its ``(lower,
        upper)`` bounds. The parameter values will be uniformly
        distributed between these bounds, inclusively.

    seed : int, optional
        A seed to initialize the `numpy.random.BitGenerator`. If `None`,
        then fresh, unpredictable entropy will be pulled from the OS.

    Returns
    -------
    table : `~astropy.table.QTable`
        A table of parameters for the randomly generated sources. Each
        row of the table corresponds to a source whose model parameters
        are defined by the column names. The column names will be the
        keys of the dictionary ``param_ranges``. The table will also
        contain an ``'id'`` column with unique source IDs.

    Notes
    -----
    To generate identical parameter values from separate function
    calls, ``param_ranges`` must have the same parameter ranges and the
    ``seed`` must be the same.

    Examples
    --------
    >>> from photutils.datasets import make_random_models_table
    >>> n_sources = 5
    >>> param_ranges = {'amplitude': [500, 1000],
    ...                 'x_mean': [0, 500],
    ...                 'y_mean': [0, 300],
    ...                 'x_stddev': [1, 5],
    ...                 'y_stddev': [1, 5],
    ...                 'theta': [0, np.pi]}
    >>> params = make_random_models_table(n_sources, param_ranges, seed=0)
    >>> for col in params.colnames:
    ...     params[col].info.format = '%.8g'  # for consistent table output
    >>> print(params)
     id amplitude   x_mean    y_mean    x_stddev  y_stddev   theta
    --- --------- --------- ---------- --------- --------- ---------
      1 818.48084 456.37779  244.75607 1.7026225 1.1132787 1.2053586
      2 634.89336 303.31789 0.82155005 4.4527157 1.4971331 3.1328274
      3 520.48676 364.74828  257.22128 3.1658449 3.6824977 3.0813851
      4 508.26382  271.8125  10.075673 2.1988476  3.588758 2.1536937
      5 906.63512 467.53621  218.89663 2.6907489 3.4615404 2.0434781
    """
    rng = np.random.default_rng(seed)

    sources = QTable()
    sources.meta.update(_get_meta())  # keep sources.meta type
    sources['id'] = np.arange(n_sources) + 1
    for param_name, (lower, upper) in param_ranges.items():
        # Generate a column for every item in param_ranges, even if it
        # is not in the model (e.g., flux).
        sources[param_name] = rng.uniform(lower, upper, n_sources)

    return sources


def params_table_to_models(params_table, model):
    """
    Create a list of models from a table of model parameters.

    Parameters
    ----------
    params_table : `~astropy.table.Table`
        A table containing the model parameters for each source.
        Each row of the table corresponds to a different model whose
        parameters are defined by the column names. Model parameters not
        defined in the table will be set to the ``model`` default value.
        To attach units to model parameters, ``params_table`` must be
        input as a `~astropy.table.QTable`. A column named 'name' can
        also be included in the table to assign a name to each model.

    model : `astropy.modeling.Model`
        The model whose parameters will be updated.

    Returns
    -------
    models : list of `astropy.modeling.Model`
        A list of models created from the input table of model
        parameters.

    Examples
    --------
    >>> from astropy.table import QTable
    >>> from photutils.datasets import params_table_to_models
    >>> from photutils.psf import CircularGaussianPSF
    >>> tbl = QTable()
    >>> tbl['x_0'] = [1, 2, 3]
    >>> tbl['y_0'] = [4, 5, 6]
    >>> tbl['flux'] = [100, 200, 300]
    >>> model = CircularGaussianPSF()
    >>> models = params_table_to_models(tbl, model)
    >>> models
    [<CircularGaussianPSF(flux=100., x_0=1., y_0=4., fwhm=1.)>,
     <CircularGaussianPSF(flux=200., x_0=2., y_0=5., fwhm=1.)>,
     <CircularGaussianPSF(flux=300., x_0=3., y_0=6., fwhm=1.)>]
    """
    param_names = set(model.param_names)
    colnames = set(params_table.colnames)
    if param_names.isdisjoint(colnames):
        msg = 'No matching model parameter names found in params_table'
        raise ValueError(msg)

    param_names = [*list(param_names), 'name']
    models = []
    for row in params_table:
        new_model = model.copy()
        for param_name in param_names:
            if param_name not in colnames:
                continue
            setattr(new_model, param_name, row[param_name])
        models.append(new_model)

    return models



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the core module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.stats import SigmaClip
from numpy.testing import assert_allclose, assert_equal

from photutils.background.core import (BiweightLocationBackground,
                                       BiweightScaleBackgroundRMS,
                                       MADStdBackgroundRMS, MeanBackground,
                                       MedianBackground, MMMBackground,
                                       ModeEstimatorBackground,
                                       SExtractorBackground, StdBackgroundRMS)
from photutils.datasets import make_noise_image
from photutils.utils._stats import nanmean

# Parametrize lists for background estimator classes
BACKGROUND_CLASSES = [MeanBackground, MedianBackground,
                      ModeEstimatorBackground, MMMBackground,
                      SExtractorBackground, BiweightLocationBackground]

# Parametrize lists for background RMS estimator classes
BACKGROUND_RMS_CLASSES = [StdBackgroundRMS, MADStdBackgroundRMS,
                          BiweightScaleBackgroundRMS]


@pytest.fixture
def bkg_value():
    """
    Expected background value for test data.
    """
    return 12.4


@pytest.fixture
def std_value():
    """
    Expected standard deviation for test data.
    """
    return 0.42


@pytest.fixture
def test_data(bkg_value, std_value):
    """
    Create test data with Gaussian noise.
    """
    return make_noise_image((100, 100), distribution='gaussian',
                            mean=bkg_value, stddev=std_value, seed=0)


@pytest.fixture
def sigma_clip():
    """
    Sigma clipping instance for tests.
    """
    return SigmaClip(sigma=3.0)


class TestBackgroundEstimators:
    """
    Test suite for all background estimator classes.

    Tests common functionality shared by all background estimators
    including basic calculations, axis handling, masking, units, and
    error handling.
    """

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_constant_background(self, bkg_class, sigma_clip):
        """
        Test background calculation on constant data.
        """
        data = np.ones((100, 100))
        bkg = bkg_class(sigma_clip=sigma_clip)
        bkgval = bkg.calc_background(data)
        assert not np.ma.isMaskedArray(bkgval)
        assert_allclose(bkgval, 1.0)
        assert_allclose(bkg(data), bkg.calc_background(data))

        # Test with masked array
        mask = np.zeros(data.shape, dtype=bool)
        mask[0, 0:10] = True
        data = np.ma.MaskedArray(data, mask=mask)
        bkgval = bkg.calc_background(data)
        assert not np.ma.isMaskedArray(bkgval)
        assert_allclose(bkgval, 1.0)
        assert_allclose(bkg(data), bkg.calc_background(data))

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_background_with_sigma_clip(self, bkg_class, sigma_clip,
                                        test_data, bkg_value):
        """
        Test background calculation with sigma clipping.
        """
        bkg = bkg_class(sigma_clip=sigma_clip)
        bkgval = bkg.calc_background(test_data)
        assert not np.ma.isMaskedArray(bkgval)
        assert_allclose(bkgval, bkg_value, atol=0.017)
        assert_allclose(bkg(test_data), bkg.calc_background(test_data))

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_background_without_sigma_clip(self, bkg_class, test_data,
                                           bkg_value):
        """
        Test background calculation without sigma clipping.
        """
        bkg = bkg_class(sigma_clip=None)
        bkgval = bkg.calc_background(test_data)
        assert not np.ma.isMaskedArray(bkgval)
        assert_allclose(bkgval, bkg_value, atol=0.017)
        assert_allclose(bkg(test_data), bkg.calc_background(test_data))

        # Test with masked array
        mask = np.zeros(test_data.shape, dtype=bool)
        mask[0, 0:10] = True
        data = np.ma.MaskedArray(test_data, mask=mask)
        bkgval = bkg.calc_background(data)
        assert not np.ma.isMaskedArray(bkgval)
        assert_allclose(bkgval, bkg_value, atol=0.018)
        assert_allclose(bkg(data), bkg.calc_background(data))

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_axis_parameter(self, bkg_class, sigma_clip, test_data):
        """
        Test background calculation along specific axes.
        """
        bkg = bkg_class(sigma_clip=sigma_clip)

        # Test axis=0
        bkg_arr = bkg.calc_background(test_data, axis=0)
        bkgi = np.array([bkg.calc_background(test_data[:, i])
                        for i in range(100)])
        assert_allclose(bkg_arr, bkgi)

        # Test axis=1
        bkg_arr = bkg.calc_background(test_data, axis=1)
        bkgi = []
        for i in range(100):
            bkgi.append(bkg.calc_background(test_data[i, :]))
        bkgi = np.array(bkgi)
        assert_allclose(bkg_arr, bkgi)

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_axis_tuple(self, bkg_class, test_data):
        """
        Test that axis=None and axis=(0,1) give same result.
        """
        bkg = bkg_class(sigma_clip=None)
        bkg_val1 = bkg.calc_background(test_data, axis=None)
        bkg_val2 = bkg.calc_background(test_data, axis=(0, 1))
        assert_allclose(bkg_val1, bkg_val2)

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_multidimensional_arrays(self, bkg_class):
        """
        Test background calculation with multidimensional arrays.
        """
        data1 = np.ones((1, 100, 100))
        data2 = np.ones((1, 100 * 100))
        data3 = np.ones((1, 1, 100 * 100))
        data4 = np.ones((1, 1, 1, 100 * 100))

        bkg = bkg_class(sigma_clip=None)
        val = bkg(data1, axis=None)
        assert np.ndim(val) == 0
        val = bkg(data1, axis=(1, 2))
        assert val.shape == (1,)
        val = bkg(data1, axis=-1)
        assert val.shape == (1, 100)
        val = bkg(data2, axis=-1)
        assert val.shape == (1,)
        val = bkg(data3, axis=-1)
        assert val.shape == (1, 1)
        val = bkg(data4, axis=-1)
        assert val.shape == (1, 1, 1)
        val = bkg(data4, axis=(2, 3))
        assert val.shape == (1, 1)
        val = bkg(data4, axis=(1, 2, 3))
        assert val.shape == (1,)
        val = bkg(data4, axis=(0, 1, 2))
        assert val.shape == (10000,)

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_masked_arrays(self, bkg_class, test_data, bkg_value):
        """
        Test masked array handling with masked parameter.
        """
        bkg = bkg_class(sigma_clip=None)
        mask = np.zeros(test_data.shape, dtype=bool)
        mask[0, 0:10] = True
        data = np.ma.MaskedArray(test_data, mask=mask)

        # Test masked array with masked=True with axis
        bkgval1 = bkg(data, masked=True, axis=1)
        bkgval2 = bkg.calc_background(data, masked=True, axis=1)
        assert np.ma.isMaskedArray(bkgval1)
        assert_allclose(np.mean(bkgval1), np.mean(bkgval2))
        assert_allclose(np.mean(bkgval1), bkg_value, atol=0.004)

        # Test masked array with masked=False with axis
        bkgval2 = bkg.calc_background(data, masked=False, axis=1)
        assert not np.ma.isMaskedArray(bkgval2)
        assert_allclose(nanmean(bkgval2), bkg_value, atol=0.004)

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_units(self, bkg_class, sigma_clip):
        """
        Test that units are properly propagated.
        """
        data = np.ones((100, 100)) << u.Jy
        bkg = bkg_class(sigma_clip=sigma_clip)
        bkgval = bkg.calc_background(data)
        assert isinstance(bkgval, u.Quantity)

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_repr(self, bkg_class):
        """
        Test string representation.
        """
        bkg = bkg_class()
        bkg_repr = repr(bkg)
        assert bkg_repr == str(bkg)
        assert bkg_repr.startswith(f'{bkg.__class__.__name__}')

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_invalid_sigma_clip(self, bkg_class):
        """
        Test error handling for invalid sigma_clip parameter.
        """
        match = 'sigma_clip must be an astropy SigmaClip instance or None'
        with pytest.raises(TypeError, match=match):
            bkg_class(sigma_clip=3)


class TestBackgroundEdgeCases:
    """
    Test edge cases for background estimators.

    Tests unusual data conditions like all-NaN arrays, various data
    types, 1D arrays, and completely masked arrays.
    """

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_all_nan_data(self, bkg_class):
        """
        Test behavior with all-NaN data.
        """
        data = np.full((10, 10), np.nan)
        bkg = bkg_class(sigma_clip=None)
        result = bkg.calc_background(data)
        assert np.isnan(result)

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    @pytest.mark.parametrize('dtype', [np.int8, np.int16, np.int32, np.int64,
                                       np.float16, np.float32, np.float64])
    def test_various_dtypes(self, bkg_class, dtype):
        """
        Test with various numeric data types.
        """
        data = np.ones((50, 50), dtype=dtype)
        bkg = bkg_class(sigma_clip=None)
        result = bkg.calc_background(data)
        assert_allclose(result, 1.0)

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_1d_arrays(self, bkg_class):
        """
        Test with 1D arrays.
        """
        rng = np.random.default_rng(0)
        data = rng.standard_normal(1000) + 5.0
        bkg = bkg_class(sigma_clip=None)
        result = bkg.calc_background(data)
        assert_allclose(result, 5.0, atol=0.2)

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_completely_masked_array(self, bkg_class):
        """
        Test with completely masked array.
        """
        data = np.ma.array(np.ones((10, 10)), mask=True)
        bkg = bkg_class(sigma_clip=None)
        result = bkg.calc_background(data)
        assert np.isnan(result)

    def test_sigma_clipping_effect(self):
        """
        Test that sigma clipping removes outliers for MeanBackground.
        """
        data = np.ones((100, 100))
        data[0:5, 0:5] = 1000  # Add outliers

        bkg_no_clip = MeanBackground(sigma_clip=None)
        bkg_with_clip = MeanBackground(sigma_clip=SigmaClip(sigma=3.0))

        result_no_clip = bkg_no_clip.calc_background(data)
        result_with_clip = bkg_with_clip.calc_background(data)

        # With clipping should be closer to 1.0
        assert abs(result_with_clip - 1.0) < abs(result_no_clip - 1.0)
        assert result_no_clip > 1.0
        assert_allclose(result_with_clip, 1.0, atol=0.01)


class TestSpecificEstimators:
    """
    Test class-specific behavior for individual estimators.

    Tests functionality unique to specific estimator implementations.
    """

    def test_sextractor_zero_std(self):
        """
        Test SExtractorBackground with zero standard deviation.
        """
        data = np.ones((100, 100))
        bkg = SExtractorBackground(sigma_clip=None)
        assert_allclose(bkg.calc_background(data), 1.0)

    def test_sextractor_skewed_data(self):
        """
        Test SExtractorBackground with highly skewed data.
        """
        data = np.arange(100)
        data[70:] = 1.0e7
        bkg = SExtractorBackground(sigma_clip=None)
        assert_allclose(bkg.calc_background(data), np.median(data))

    @pytest.mark.parametrize(('median_factor', 'mean_factor'), [
        (3.0, 2.0),  # default
        (2.5, 1.5),  # custom
        (1.0, 0.0),  # edge case
    ])
    def test_mode_estimator_parameters(self, median_factor, mean_factor):
        """
        Test ModeEstimatorBackground with different parameters.
        """
        rng = np.random.default_rng(0)
        data = rng.standard_normal((100, 100)) + 10
        bkg = ModeEstimatorBackground(median_factor=median_factor,
                                      mean_factor=mean_factor,
                                      sigma_clip=None)
        result = bkg.calc_background(data)

        # Verify the formula is applied correctly
        expected = (median_factor * np.median(data)
                    - mean_factor * np.mean(data))
        assert_allclose(result, expected)

    def test_biweight_location_constant_data(self):
        """
        Test BiweightLocationBackground with constant data.

        Regression test for
        https://github.com/astropy/astropy/pull/16964.
        """
        data = np.full((5, 1, 10), 7.5)
        bkg = BiweightLocationBackground(sigma_clip=None)
        result = bkg.calc_background(data, axis=-1)
        assert result.shape == (5, 1)

        data = np.full((5, 1, 4, 10), 1.4)
        bkg = BiweightLocationBackground(sigma_clip=None)
        result = bkg.calc_background(data, axis=-1)
        assert result.shape == (5, 1, 4)

        data = np.full((5, 1, 10), 1.4)
        bkg = BiweightLocationBackground(sigma_clip=None)
        result1 = bkg.calc_background(data, axis=None)
        result2 = bkg.calc_background(data, axis=(0, 1, 2))
        assert result1.shape == ()
        assert result2.shape == ()
        assert result1 == result2

    def test_biweight_scale_constant_data(self):
        """
        Test BiweightScaleBackgroundRMS with constant data.

        Regression test for
        https://github.com/astropy/astropy/pull/16964.
        """
        data = np.full((5, 1, 10), 7.5)
        bkg = BiweightScaleBackgroundRMS(sigma_clip=None)
        result = bkg.calc_background_rms(data, axis=-1)
        assert result.shape == (5, 1)

        data = np.full((5, 1, 4, 10), 1.4)
        bkg = BiweightScaleBackgroundRMS(sigma_clip=None)
        result = bkg.calc_background_rms(data, axis=-1)
        assert result.shape == (5, 1, 4)

        data = np.full((5, 1, 10), 1.4)
        bkg = BiweightScaleBackgroundRMS(sigma_clip=None)
        result1 = bkg.calc_background_rms(data, axis=None)
        result2 = bkg.calc_background_rms(data, axis=(0, 1, 2))
        assert result1.shape == ()
        assert result2.shape == ()
        assert result1 == result2


class TestBackgroundRMSEstimators:
    """
    Test suite for all background RMS estimator classes.

    Tests common functionality shared by all RMS estimators.
    """

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_background_rms_with_sigma_clip(self, rms_class, sigma_clip,
                                            test_data, std_value):
        """
        Test RMS calculation with sigma clipping.
        """
        bkgrms = rms_class(sigma_clip=sigma_clip)
        assert_allclose(bkgrms.calc_background_rms(test_data), std_value,
                        atol=0.007)
        assert_allclose(bkgrms(test_data),
                        bkgrms.calc_background_rms(test_data))

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_background_rms_without_sigma_clip(self, rms_class, test_data,
                                               std_value):
        """
        Test RMS calculation without sigma clipping.
        """
        bkgrms = rms_class(sigma_clip=None)
        assert_allclose(bkgrms.calc_background_rms(test_data), std_value,
                        atol=0.004)
        assert_allclose(bkgrms(test_data),
                        bkgrms.calc_background_rms(test_data))

        # Test with masked array
        mask = np.zeros(test_data.shape, dtype=bool)
        mask[0, 0:10] = True
        data = np.ma.MaskedArray(test_data, mask=mask)
        rms = bkgrms.calc_background_rms(data)
        assert not np.ma.isMaskedArray(bkgrms)
        assert_allclose(rms, std_value, atol=0.004)
        assert_allclose(bkgrms(data), bkgrms.calc_background_rms(data))

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_axis_parameter(self, rms_class, sigma_clip, test_data):
        """
        Test RMS calculation along specific axes.
        """
        bkgrms = rms_class(sigma_clip=sigma_clip)

        # Test axis=0
        rms_arr = bkgrms.calc_background_rms(test_data, axis=0)
        rmsi = np.array([bkgrms.calc_background_rms(test_data[:, i])
                        for i in range(100)])
        assert_allclose(rms_arr, rmsi)

        # Test axis=1
        rms_arr = bkgrms.calc_background_rms(test_data, axis=1)
        rmsi = []
        for i in range(100):
            rmsi.append(bkgrms.calc_background_rms(test_data[i, :]))
        rmsi = np.array(rmsi)
        assert_allclose(rms_arr, rmsi)

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_multidimensional_arrays(self, rms_class):
        """
        Test RMS calculation with multidimensional arrays.
        """
        data1 = np.ones((1, 100, 100))
        data2 = np.ones((1, 100 * 100))
        data3 = np.ones((1, 1, 100 * 100))
        data4 = np.ones((1, 1, 1, 100 * 100))

        bkgrms = rms_class(sigma_clip=None)
        val = bkgrms(data1, axis=None)
        assert np.ndim(val) == 0
        val = bkgrms(data1, axis=(1, 2))
        assert val.shape == (1,)
        val = bkgrms(data1, axis=-1)
        assert val.shape == (1, 100)
        val = bkgrms(data2, axis=-1)
        assert val.shape == (1,)
        val = bkgrms(data3, axis=-1)
        assert val.shape == (1, 1)
        val = bkgrms(data4, axis=-1)
        assert val.shape == (1, 1, 1)
        val = bkgrms(data4, axis=(2, 3))
        assert val.shape == (1, 1)
        val = bkgrms(data4, axis=(1, 2, 3))
        assert val.shape == (1,)
        val = bkgrms(data4, axis=(0, 1, 2))
        assert val.shape == (10000,)

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_masked_arrays(self, rms_class, test_data, std_value):
        """
        Test masked array handling with masked parameter.
        """
        bkgrms = rms_class(sigma_clip=None)
        mask = np.zeros(test_data.shape, dtype=bool)
        mask[0, 0:10] = True
        data = np.ma.MaskedArray(test_data, mask=mask)

        # Test masked array with masked=True with axis
        rms1 = bkgrms(data, masked=True, axis=1)
        rms2 = bkgrms.calc_background_rms(data, masked=True, axis=1)
        assert np.ma.isMaskedArray(rms1)
        assert_allclose(np.mean(rms1), np.mean(rms2))
        assert_allclose(np.mean(rms1), std_value, atol=0.04)

        # Test masked array with masked=False with axis
        rms3 = bkgrms.calc_background_rms(data, masked=False, axis=1)
        assert not np.ma.isMaskedArray(rms3)
        assert_allclose(nanmean(rms3), std_value, atol=0.04)

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_units(self, rms_class, sigma_clip):
        """
        Test that units are properly propagated.
        """
        data = np.ones((100, 100)) << u.Jy
        bkgrms = rms_class(sigma_clip=sigma_clip)
        rmsval = bkgrms.calc_background_rms(data)
        assert isinstance(rmsval, u.Quantity)

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_repr(self, rms_class):
        """
        Test string representation.
        """
        bkgrms = rms_class()
        rms_repr = repr(bkgrms)
        assert rms_repr == str(bkgrms)
        assert rms_repr.startswith(f'{bkgrms.__class__.__name__}')

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_invalid_sigma_clip(self, rms_class):
        """
        Test error handling for invalid sigma_clip parameter.
        """
        match = 'sigma_clip must be an astropy SigmaClip instance or None'
        with pytest.raises(TypeError, match=match):
            rms_class(sigma_clip=3)


class TestInputNotMutated:
    """
    Test that input data is never modified by background estimator
    calls.
    """

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_background_does_not_mutate_input(self, bkg_class):
        """
        Test that calc_background does not modify a plain ndarray.
        """
        data = np.ones((100, 100))
        data[0:5, 0:5] = 1000.0  # outliers that trigger sigma clipping
        data_orig = data.copy()

        bkg = bkg_class(sigma_clip=SigmaClip(sigma=3.0))
        bkg.calc_background(data)
        assert_equal(data, data_orig)

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_background_does_not_mutate_input_no_clip(self, bkg_class):
        """
        Test that calc_background does not modify input when sigma
        clipping is disabled.
        """
        data = np.ones((100, 100))
        data[0:5, 0:5] = 1000.0
        data_orig = data.copy()

        bkg = bkg_class(sigma_clip=None)
        bkg.calc_background(data)
        assert_equal(data, data_orig)

    @pytest.mark.parametrize('bkg_class', BACKGROUND_CLASSES)
    def test_background_does_not_mutate_masked_input(self, bkg_class):
        """
        Test that calc_background does not modify a masked array (data
        values or mask).
        """
        data = np.ones((100, 100))
        data[0:5, 0:5] = 1000.0
        mask = np.zeros(data.shape, dtype=bool)
        mask[0:5, 0:5] = True
        data_ma = np.ma.MaskedArray(data.copy(), mask=mask)
        data_values_orig = data_ma.data.copy()
        mask_orig = mask.copy()

        bkg = bkg_class(sigma_clip=None)
        bkg.calc_background(data_ma)
        assert_equal(data_ma.data, data_values_orig)
        assert_equal(data_ma.mask, mask_orig)

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_background_rms_does_not_mutate_input(self, rms_class):
        """
        Test that calc_background_rms does not modify a plain ndarray.
        """
        data = np.ones((100, 100))
        data[0:5, 0:5] = 1000.0
        data_orig = data.copy()

        bkgrms = rms_class(sigma_clip=SigmaClip(sigma=3.0))
        bkgrms.calc_background_rms(data)
        assert_equal(data, data_orig)

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_background_rms_does_not_mutate_input_no_clip(self, rms_class):
        """
        Test that calc_background_rms does not modify input when sigma
        clipping is disabled.
        """
        data = np.ones((100, 100))
        data[0:5, 0:5] = 1000.0
        data_orig = data.copy()

        bkgrms = rms_class(sigma_clip=None)
        bkgrms.calc_background_rms(data)
        assert_equal(data, data_orig)

    @pytest.mark.parametrize('rms_class', BACKGROUND_RMS_CLASSES)
    def test_background_rms_does_not_mutate_masked_input(self, rms_class):
        """
        Test that calc_background_rms does not modify a masked array
        (data values or mask).
        """
        data = np.ones((100, 100))
        data[0:5, 0:5] = 1000.0
        mask = np.zeros(data.shape, dtype=bool)
        mask[0:5, 0:5] = True
        data_ma = np.ma.MaskedArray(data.copy(), mask=mask)
        data_values_orig = data_ma.data.copy()
        mask_orig = mask.copy()

        bkgrms = rms_class(sigma_clip=None)
        bkgrms.calc_background_rms(data_ma)
        assert_equal(data_ma.data, data_values_orig)
        assert_equal(data_ma.mask, mask_orig)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for estimating the 2D background and background RMS in an image.
"""

import copy
import warnings

import astropy.units as u
import numpy as np
from astropy.nddata import NDData, block_replicate, reshape_as_blocks
from astropy.utils import lazyproperty
from astropy.utils.exceptions import AstropyUserWarning
from scipy.ndimage import generic_filter

from photutils.aperture import RectangularAperture
from photutils.background.core import (SIGMA_CLIP, SExtractorBackground,
                                       StdBackgroundRMS)
from photutils.background.interpolators import (BkgIDWInterpolator,
                                                _BkgZoomInterpolator)
from photutils.utils import ShepardIDWInterpolator
from photutils.utils._deprecation import (deprecated,
                                          deprecated_renamed_argument)
from photutils.utils._parameters import as_pair, create_default_sigmaclip
from photutils.utils._repr import make_repr
from photutils.utils._stats import nanmedian, nanmin

__all__ = ['Background2D']

__doctest_skip__ = ['Background2D']


class Background2D:
    """
    Class to estimate a 2D background and background RMS noise in an
    image.

    The background is estimated using (sigma-clipped) statistics in
    each box of a grid that covers the input ``data`` to create a
    low-resolution, and possibly irregularly-gridded, background map.

    The final background map is calculated by interpolating the
    low-resolution background map.

    Invalid data values (i.e., NaN or inf) are automatically masked.

    .. note::

        Better performance will generally be obtained if you have the
        `Bottleneck <https://github.com/pydata/bottleneck>`_ package
        installed. See :ref:`performance-tips` for details, including
        notes on array byte order (endianness) when loading FITS data.

    Parameters
    ----------
    data : array_like or `~astropy.nddata.NDData`
        The 2D array from which to estimate the background and/or
        background RMS map.

    box_size : int or array_like (int)
        The box size along each axis. If ``box_size`` is a scalar then
        a square box of size ``box_size`` will be used. If ``box_size``
        has two elements, they must be in ``(ny, nx)`` order. For best
        results, the box shape should be chosen such that the ``data``
        are covered by an integer number of boxes in both dimensions.
        When this is not the case, the image will be padded along the
        top and/or right edges. Ideally, the ``box_size`` should be
        chosen such that an integer number of boxes is only slightly
        larger than the ``data`` size to minimize the amount of padding.

    mask : array_like (bool), optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        Masked data are excluded from the background and background
        RMS calculations. ``mask`` is intended to mask sources or bad
        pixels, but a background and background RMS value will be
        calculated for them based on interpolation of the low-resolution
        background and background RMS maps. Use ``coverage_mask`` to
        mask blank areas of an image. ``coverage_mask`` pixels are
        assigned a value of ``fill_value`` (default = 0) in the output
        background and background RMS maps.

    coverage_mask : array_like (bool), optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        ``coverage_mask`` should be `True` where there is no coverage
        (i.e., no data) for a given pixel (e.g., blank areas in a mosaic
        image). It should not be used to mask sources or bad pixels (in
        that case use ``mask`` instead). ``coverage_mask`` pixels are
        assigned a value of ``fill_value`` (default = 0) in the output
        background and background RMS maps.

    fill_value : float, optional
        The value used to fill the output background and background RMS
        maps where the input ``coverage_mask`` is `True`.

    exclude_percentile : float in the range of [0, 100], optional
        The percentage of masked pixels allowed in a box for it to be
        included in the low-resolution map. If a box has more than
        ``exclude_percentile`` percent of its pixels masked then it
        will be excluded from the low-resolution map. Masked pixels
        include those from the input ``mask`` and ``coverage_mask``,
        non-finite ``data`` values, any padded area at the data
        edges, and those resulting from any sigma clipping. Setting
        ``exclude_percentile=0`` will exclude boxes that have any that
        have any masked pixels. Note that completely masked boxes are
        always excluded. In general, ``exclude_percentile`` should be
        kept as low as possible to ensure there are a sufficient number
        of unmasked pixels in each box for reasonable statistical
        estimates. The default is 10.0.

    filter_size : int or array_like (int), optional
        The window size of the 2D median filter to apply to the
        low-resolution background map. If ``filter_size`` is a scalar
        then a square box of size ``filter_size`` will be used. If
        ``filter_size`` has two elements, they must be in ``(ny, nx)``
        order. ``filter_size`` must be odd along both axes. A filter
        size of ``1`` (or ``(1, 1)``) means no filtering.

    filter_threshold : int, optional
        The threshold value for used for selective median filtering of
        the low-resolution 2D background map. The median filter will
        be applied to only the background boxes with values larger
        than ``filter_threshold``. Set to `None` to filter all boxes
        (default).

    sigma_clip : `astropy.stats.SigmaClip` or `None`, optional
        A `~astropy.stats.SigmaClip` object that defines the sigma
        clipping parameters. If `None` then no sigma clipping will
        be performed.

    bkg_estimator : callable, optional
        A callable object (a function or e.g., an instance of
        any `~photutils.background.BackgroundBase` subclass)
        used to estimate the background in each of the boxes.
        The callable object must take in a 2D `~numpy.ndarray`
        or `~numpy.ma.MaskedArray` and have an ``axis`` keyword.
        Internally, the background will be calculated along ``axis=1``
        and in this case the callable object must return a 1D
        `~numpy.ndarray`, where np.nan values are used for masked
        pixels. If ``bkg_estimator`` includes sigma clipping, it
        will be ignored (use the ``sigma_clip`` keyword here to
        define sigma clipping). The default is an instance of
        `~photutils.background.SExtractorBackground`.

    bkg_rms_estimator : callable, optional
        A callable object (a function or e.g., an instance of
        any `~photutils.background.BackgroundRMSBase` subclass)
        used to estimate the background RMS in each of the boxes.
        The callable object must take in a 2D `~numpy.ndarray`
        or `~numpy.ma.MaskedArray` and have an ``axis`` keyword.
        Internally, the background RMS will be calculated along
        ``axis=1`` and in this case the callable object must return a
        1D `~numpy.ndarray`, where np.nan values are used for masked
        pixels. If ``bkg_rms_estimator`` includes sigma clipping,
        it will be ignored (use the ``sigma_clip`` keyword here
        to define sigma clipping). The default is an instance of
        `~photutils.background.StdBackgroundRMS`.

    interpolator : callable, optional
        A callable object (a function or object) used to interpolate
        the low-resolution background or background RMS image to the
        full-size background or background RMS maps. The default
        is an instance of `BkgZoomInterpolator`, which uses the
        `scipy.ndimage.zoom` function.

        .. deprecated:: 3.0
           This keyword argument is deprecated and will be removed in
           a future version. When removed, the `scipy.ndimage.zoom`
           cubic spline interpolator will always be used to resize the
           low-resolution background and background RMS arrays to the
           full image size.

    Notes
    -----
    Integer input data produce background and background RMS outputs
    with ``np.float32`` dtype to preserve precision from interpolation
    while minimizing memory usage. Float input data produce background
    and background RMS outputs with the same dtype as the input data.

    If there is only one background box element (i.e., ``box_size`` is
    the same size as (or larger than) the ``data``), then the background
    map will simply be a constant image.
    """

    @deprecated_renamed_argument('bkgrms_estimator', 'bkg_rms_estimator',
                                 '3.0', until='4.0')
    @deprecated_renamed_argument('interpolator', None, '3.0', until='4.0')
    def __init__(self, data, box_size, *, mask=None, coverage_mask=None,
                 fill_value=0.0, exclude_percentile=10.0, filter_size=(3, 3),
                 filter_threshold=None, sigma_clip=SIGMA_CLIP,
                 bkg_estimator=None, bkg_rms_estimator=None,
                 interpolator=None):

        if isinstance(data, (u.Quantity, NDData)):  # includes CCDData
            self._unit = data.unit
            data = data.data
        else:
            self._unit = None

        # self._data is a temporary instance variable to store the input
        # data (the variable is deleted in self._calculate_stats)
        self._data = self._validate_array(data, 'data', shape=False)
        self._data_dtype = self._data.dtype
        self._data_shape = self._data.shape
        if np.all(~np.isfinite(self._data)):
            msg = ('Input data contains all non-finite (NaN or infinity) '
                   'values. Cannot compute a background.')
            raise ValueError(msg)

        # self._mask is a temporary instance variable to store the input
        # mask array (deleted in self._calculate_stats); self._has_mask
        # records whether a mask was provided for use after deletion.
        self._mask = self._validate_array(mask, 'mask')
        self._has_mask = self._mask is not None
        self.coverage_mask = self._validate_array(coverage_mask,
                                                  'coverage_mask')

        # box_size cannot be larger than the data array size
        self.box_size = as_pair('box_size', box_size, lower_bound=(0, 0),
                                upper_bound=data.shape)

        self.fill_value = fill_value
        if exclude_percentile < 0 or exclude_percentile > 100:
            msg = 'exclude_percentile must be between 0 and 100 (inclusive)'
            raise ValueError(msg)
        self.exclude_percentile = exclude_percentile
        self.filter_size = as_pair('filter_size', filter_size,
                                   lower_bound=(0, 0), check_odd=True)
        self.filter_threshold = filter_threshold

        if sigma_clip is SIGMA_CLIP:
            sigma_clip = create_default_sigmaclip(sigma=SIGMA_CLIP.sigma,
                                                  maxiters=SIGMA_CLIP.maxiters)
        self.sigma_clip = sigma_clip

        if interpolator is None:
            interpolator = _BkgZoomInterpolator()
        self.interpolator = interpolator

        if bkg_estimator is None:
            bkg_estimator = SExtractorBackground(sigma_clip=None)
        if bkg_rms_estimator is None:
            bkg_rms_estimator = StdBackgroundRMS(sigma_clip=None)

        # We perform sigma clipping as a separate step to avoid calling
        # it twice for the background and background RMS. Shallow-copy
        # the estimators before mutating their sigma_clip attribute so
        # that any user-supplied estimator object is not modified in
        # place.
        bkg_estimator = copy.copy(bkg_estimator)
        bkg_rms_estimator = copy.copy(bkg_rms_estimator)
        if hasattr(bkg_estimator, 'sigma_clip'):
            bkg_estimator.sigma_clip = None
        if hasattr(bkg_rms_estimator, 'sigma_clip'):
            bkg_rms_estimator.sigma_clip = None
        self.bkg_estimator = bkg_estimator
        self.bkg_rms_estimator = bkg_rms_estimator

        self._box_npixels = None

        # Store the interpolator keyword arguments for later use
        # (before self._data is deleted in self._calculate_stats)
        interp_dtype = self._data.dtype
        if interp_dtype.kind != 'f':
            interp_dtype = np.float32
        self._interp_kwargs = {'shape': self._data.shape,
                               'dtype': interp_dtype,
                               'box_size': self.box_size}

        # Perform the initial calculations to avoid storing large data
        # arrays and to keep the memory usage minimal
        (self._bkg_stats,
         self._bkgrms_stats,
         self._ngood) = self._calculate_stats()

        # This is used to selectively filter the low-resolution maps
        self._min_bkg_stats = nanmin(self._bkg_stats)

        # Store a mask of the excluded mesh values (NaNs) in the
        # low-resolution maps
        self._mesh_nan_mask = np.isnan(self._bkg_stats)

        # Add keyword arguments needed for BkgZoomInterpolator.
        # BkgIDWInterpolator upscales the mesh based only on the good
        # pixels in the low-resolution mesh.
        if isinstance(self.interpolator, BkgIDWInterpolator):
            self._interp_kwargs['mesh_yxcen'] = self._calculate_mesh_yxcen()
            self._interp_kwargs['mesh_nan_mask'] = self._mesh_nan_mask

    def _repr_str_params(self):
        params = ('data', 'box_size', 'mask', 'coverage_mask', 'fill_value',
                  'exclude_percentile', 'filter_size', 'filter_threshold',
                  'sigma_clip', 'bkg_estimator', 'bkg_rms_estimator',
                  'interpolator')

        data_repr = f'<array; shape={self._interp_kwargs["shape"]}>'

        mask_repr = None if not self._has_mask else data_repr

        if 'coverage_mask' in self.__dict__ and self.coverage_mask is None:
            coverage_mask_repr = None
        else:
            coverage_mask_repr = data_repr

        overrides = {'data': data_repr, 'mask': mask_repr,
                     'coverage_mask': coverage_mask_repr}

        return params, overrides

    def __repr__(self):
        params, overrides = self._repr_str_params()
        return make_repr(self, params, overrides=overrides)

    def __str__(self):
        params, overrides = self._repr_str_params()
        return make_repr(self, params, overrides=overrides, long=True)

    def _validate_array(self, array, name, *, shape=True):
        """
        Validate the input data, mask, and coverage_mask arrays.
        """
        if name in ('mask', 'coverage_mask') and array is np.ma.nomask:
            array = None
        if array is not None:
            array = np.asanyarray(array)
            if array.ndim != 2:
                msg = f'{name} must be a 2D array'
                raise ValueError(msg)
            if shape and array.shape != self._data.shape:
                msg = f'data and {name} must have the same shape'
                raise ValueError(msg)
        return array

    def _apply_units(self, data):
        """
        Apply units to the data.

        The units are based on the units of the input ``data`` array.

        Parameters
        ----------
        data : `~numpy.ndarray`
            The input data array.

        Returns
        -------
        data : `~numpy.ndarray`
            The data array with units applied.
        """
        if self._unit is not None:
            data <<= self._unit
        return data

    def _combine_input_masks(self):
        """
        Combine the input mask and coverage_mask.
        """
        if self._mask is None and self.coverage_mask is None:
            return None
        if self._mask is None:
            return self.coverage_mask
        if self.coverage_mask is None:
            return self._mask

        return np.logical_or(self._mask, self.coverage_mask)

    def _combine_all_masks(self, mask):
        """
        Combine the input masks (mask and coverage_mask) with the mask
        of invalid data values.
        """
        input_mask = self._combine_input_masks()

        msg = ('Input data contains non-finite (NaN or infinity) values, '
               'which were automatically masked.')

        if input_mask is None:
            if np.any(mask):
                warnings.warn(msg, AstropyUserWarning)

            total_mask = mask
        else:
            condition = np.logical_and(np.logical_not(input_mask), mask)
            if np.any(condition):
                warnings.warn(msg, AstropyUserWarning)

            total_mask = np.logical_or(input_mask, mask)

        if np.all(total_mask):
            msg = 'All input pixels are masked. Cannot compute a background.'
            raise ValueError(msg)

        return total_mask

    @lazyproperty
    def _good_npixels_threshold(self):
        """
        The minimum number of required unmasked pixels in a box used for
        it to be included in the low-resolution map.

        For exclude_percentile=0, only boxes where nmasked=0 will be
        included. For exclude_percentile=100, all boxes will be included
        *unless* they are completely masked.

        Boxes that are completely masked are always excluded.
        """
        return (1 - (self.exclude_percentile / 100.0)) * self._box_npixels

    def _sigmaclip_boxes(self, data, axis):
        """
        Sigma clip the boxes along the specified axis.

        This method sigma clips the boxes along the specified axis and
        returns the sigma-clipped data. The input ``data`` is typically
        a 4D array where the first two dimensions represent the y and x
        positions of the boxes and the last two dimensions represent the
        y and x positions within each box.

        We perform sigma clipping as a separate step to avoid performing
        sigma clipping for both the background and background RMS
        estimators.

        Parameters
        ----------
        data : `~numpy.ndarray`
            The 4D array of box data.

        axis : int or tuple of int
            The axis or axes along which to sigma clip the data.

        Returns
        -------
        data : `~numpy.ndarray`
            The sigma-clipped data.
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=AstropyUserWarning)
            if self.sigma_clip is not None:
                data = self.sigma_clip(data, axis=axis, masked=False,
                                       copy=False)

        return data

    def _compute_box_statistics(self, data, *, axis=None):
        """
        Compute the background and background RMS statistics in each
        box.

        Parameters
        ----------
        data : `~numpy.ndarray`
            The 4D array of box data.

        axis : int or tuple of int, optional
            The axis or axes along which to compute the statistics.

        Returns
        -------
        bkg : 2D `~numpy.ndarray` or float
            The background statistics in each box.

        bkgrms : 2D `~numpy.ndarray` or float
            The background RMS statistics in each box.
        """
        data = self._sigmaclip_boxes(data, axis=axis)

        # Make 2D arrays of the box statistics
        bkg = self.bkg_estimator(data, axis=axis)
        bkgrms = self.bkg_rms_estimator(data, axis=axis)

        # Mask boxes with too few unmasked pixels
        ngood = np.count_nonzero(~np.isnan(data), axis=axis)
        box_mask = ngood <= self._good_npixels_threshold

        if np.ndim(bkg) == 0:
            if box_mask:  # single corner box
                # np.nan is float64; use np.float32 to prevent numpy from
                # promoting the output data dtype to float64 if the
                # input data is float32
                bkg = np.float32(np.nan)
                bkgrms = np.float32(np.nan)
        else:
            bkg[box_mask] = np.nan
            bkgrms[box_mask] = np.nan

        return bkg, bkgrms, ngood

    def _calculate_stats(self):
        """
        Calculate the background and background RMS statistics in each
        box.

        Returns
        -------
        bkg : 2D `~numpy.ndarray`
            The background statistics in each box.

        bkgrms : 2D `~numpy.ndarray`
            The background RMS statistics in each box.

        ngood : 2D `~numpy.ndarray`
            The number of unmasked pixels in each box.
        """
        # If needed, copy the data to a float32 array to insert NaNs
        if self._data.dtype.kind != 'f':
            self._data = self._data.astype(np.float32)

        # Automatically mask non-finite values that aren't already
        # masked and combine all masks
        mask = self._combine_all_masks(~np.isfinite(self._data))

        self._box_npixels = np.prod(self.box_size)
        nboxes = self._data.shape // self.box_size
        y1, x1 = nboxes * self.box_size

        # Core boxes - the part of the data array that is an integer
        # multiple of the box size.
        # Combine the last two axes for performance.
        # Below we transform both the data and mask arrays to avoid
        # making multiple copies of the data (one to insert NaN and
        # another for the reshape). Only one copy of the data and mask
        # array is made (except for the extra corner). The boolean mask
        # copy is much smaller than the data array.
        # An explicit copy of the data array is needed to avoid
        # modifying the original data array if the shape of the data
        # array is (y1, x1) (i.e., box_size = data.shape).
        core = reshape_as_blocks(self._data[:y1, :x1].copy(), self.box_size)
        core_mask = reshape_as_blocks(mask[:y1, :x1], self.box_size)
        core = core.reshape((*nboxes, -1))
        core_mask = core_mask.reshape((*nboxes, -1))
        core[core_mask] = np.nan
        bkg, bkgrms, ngood = self._compute_box_statistics(core, axis=-1)

        extra_row = y1 < self._data.shape[0]
        extra_col = x1 < self._data.shape[1]
        if extra_row or extra_col:
            if extra_row:
                # Extra row of boxes.
                # Here we need to make a copy of the data array to avoid
                # modifying the original data array.
                # Move the axes and combine the last two for performance.
                row_data = self._data[y1:, :x1].copy()
                row_mask = mask[y1:, :x1]
                row_data[row_mask] = np.nan
                row_data = reshape_as_blocks(row_data, (1, self.box_size[1]))
                row_data = np.moveaxis(row_data, 0, -1)
                row_data = row_data.reshape((*row_data.shape[:-2], -1))
                row_bkg, row_bkgrms, row_ngood = self._compute_box_statistics(
                    row_data, axis=-1)

            if extra_col:
                # Extra column of boxes.
                # Here we need to make a copy of the data array to avoid
                # modifying the original data array.
                # Move the axes and combine the last two for performance.
                col_data = self._data[:y1, x1:].copy()
                col_mask = mask[:y1, x1:]
                col_data[col_mask] = np.nan
                col_data = reshape_as_blocks(col_data, (self.box_size[0], 1))
                col_data = np.transpose(col_data, (0, 3, 1, 2))
                col_data = col_data.reshape((*col_data.shape[:-2], -1))
                col_bkg, col_bkgrms, col_ngood = self._compute_box_statistics(
                    col_data, axis=-1)

            if extra_row and extra_col:
                # Extra corner box -- append to extra column.
                # Here we need to make a copy of the data array to avoid
                # modifying the original data array.
                corner_data = self._data[y1:, x1:].copy()
                corner_mask = mask[y1:, x1:]
                corner_data[corner_mask] = np.nan
                crn_bkg, crn_bkgrms, crn_ngood = self._compute_box_statistics(
                    corner_data, axis=None)
                col_bkg = np.vstack((col_bkg, crn_bkg))
                col_bkgrms = np.vstack((col_bkgrms, crn_bkgrms))
                col_ngood = np.vstack((col_ngood, crn_ngood))

            # Combine the core and extra boxes to construct the
            # complete 2D bkg and bkgrms arrays
            if extra_row:
                bkg = np.vstack([bkg, row_bkg[:, 0]])
                bkgrms = np.vstack([bkgrms, row_bkgrms[:, 0]])
                ngood = np.vstack([ngood, row_ngood[:, 0]])

            if extra_col:
                bkg = np.hstack([bkg, col_bkg])
                bkgrms = np.hstack([bkgrms, col_bkgrms])
                ngood = np.hstack([ngood, col_ngood])

        if np.all(np.isnan(bkg)):
            msg = (f'All boxes contain <= {self._good_npixels_threshold} '
                   f'unmasked or finite pixels ({self.box_size=}, '
                   f'{self.exclude_percentile=}). Please check your data '
                   'or increase "exclude_percentile" to allow more boxes to '
                   'be included.')
            raise ValueError(msg)

        # We no longer need the temporary input arrays
        del self._data
        del self._mask

        return bkg, bkgrms, ngood

    def _interpolate_grid(self, data, *, n_neighbors=10, eps=0.0, power=1.0,
                          regularization=0.0):
        """
        Fill in any NaN values in the low-resolution 2D mesh background
        and background RMS images using inverse distance weighting (IDW)
        interpolation.

        This method ensures that the low-resolution mesh contains no
        NaNs before applying a regular-grid interpolator to expand it to
        the full image size. If there are no NaNs, the input is returned
        (cast to the original dtype). Otherwise, NaN pixels are replaced
        by IDW interpolation using valid mesh values.

        Parameters
        ----------
        data : 2D `~numpy.ndarray`
            A 2D array of the box statistics, possibly containing NaNs.

        n_neighbors : int, optional
            The maximum number of nearest neighbors to use during the
            interpolation.

        eps : float, optional
            Approximation parameter for nearest neighbors (see
            `scipy.spatial.cKDTree.query`).

        power : float, optional
            The power of the inverse distance used for the interpolation
            weights.

        regularization : float, optional
            Regularization parameter to control the smoothness of the
            interpolator.

        Returns
        -------
        result : 2D `~numpy.ndarray`
            The input array with NaNs replaced by interpolated values.
        """
        if not np.any(np.isnan(data)):
            return data

        mask = ~np.isnan(data)
        idx = np.where(mask)
        yx = np.column_stack(idx)
        interp_func = ShepardIDWInterpolator(yx, data[mask])

        # Interpolate the masked pixels where data is NaN
        idx = np.where(np.isnan(data))
        yx_indices = np.column_stack(idx)
        interp_values = interp_func(yx_indices, n_neighbors=n_neighbors,
                                    power=power, eps=eps,
                                    regularization=regularization)

        interp_data = np.copy(data)  # copy to avoid modifying the input data
        interp_data[idx] = interp_values

        return interp_data

    def _selective_filter(self, data):
        """
        Filter only pixels above ``filter_threshold`` in a low-
        resolution 2D image.

        The pixels to be filtered are determined by applying the
        ``filter_threshold`` to the low-resolution background mesh. The
        same pixels are filtered in both the background and background
        RMS meshes.

        Parameters
        ----------
        data : 2D `~numpy.ndarray`
            A 2D array of mesh values.

        Returns
        -------
        result : 2D `~numpy.ndarray`
            The filtered 2D array of mesh values.
        """
        bkg_stats_interp = self._interpolate_grid(self._bkg_stats)
        above_threshold = bkg_stats_interp > self.filter_threshold
        if not np.any(above_threshold):
            return data

        # Apply the median filter across the whole mesh in one call,
        # then blend: keep the filtered value only where the background
        # is above the threshold; use the original value everywhere
        # else.
        filtered = generic_filter(data, nanmedian, size=self.filter_size,
                                  mode='constant', cval=np.nan)
        return np.where(above_threshold, filtered, data)

    def _filter_grid(self, data):
        """
        Apply a 2D median filter to a low-resolution 2D image.

        Parameters
        ----------
        data : 2D `~numpy.ndarray`
            A 2D array of mesh values.

        Returns
        -------
        result : 2D `~numpy.ndarray`
            The filtered 2D array of mesh values.
        """
        if tuple(self.filter_size) == (1, 1):
            return data

        if (self.filter_threshold is None
                or self.filter_threshold < self._min_bkg_stats):
            # Filter the entire array
            filtdata = generic_filter(data, nanmedian, size=self.filter_size,
                                      mode='constant', cval=np.nan)
        else:
            # Selectively filter the array
            filtdata = self._selective_filter(data)

        return filtdata

    def _calculate_mesh_yxcen(self):
        """
        Calculate the y and x positions of the centers of the low-
        resolution background and background RMS meshes with respect to
        the input data array.

        This is used by the IDW interpolator to expand the low-
        resolution mesh to the full-size image. It is also used to plot
        the mesh boxes on the input image.
        """
        mesh_idx = np.where(~self._mesh_nan_mask)  # good mesh indices
        box_cen = (self.box_size - 1) / 2.0
        return (mesh_idx * self.box_size[:, None]) + box_cen[:, None]

    def _try_free_bkg_stats(self):
        """
        Free ``_bkg_stats`` when it is safe to do so.

        ``_bkg_stats`` is always needed by ``background_mesh``
        (via ``_interpolate_grid``). It is also needed by
        ``_selective_filter`` (called from ``_filter_grid``) when
        ``filter_threshold`` is not ``None``. It is therefore
        safe to free it only after ``background_mesh`` has been
        cached and either ``filter_threshold`` is ``None``
        (so ``background_rms_mesh`` does not need it) or
        ``background_rms_mesh`` has also been cached.
        """
        if 'background_mesh' not in self.__dict__:
            return
        if (self.filter_threshold is None
                or 'background_rms_mesh' in self.__dict__):
            self._bkg_stats = None  # delete to save memory

    @lazyproperty
    def background_mesh(self):
        """
        The low-resolution background image.

        This image is equivalent to the low-resolution "MINIBACK"
        background map check image in SourceExtractor.
        """
        data = self._interpolate_grid(self._bkg_stats)
        result = self._apply_units(self._filter_grid(data))
        self._try_free_bkg_stats()
        return result

    @lazyproperty
    def background_rms_mesh(self):
        """
        The low-resolution background RMS image.

        This image is equivalent to the low-resolution "MINIBACK_RMS"
        background rms map check image in SourceExtractor.
        """
        data = self._interpolate_grid(self._bkgrms_stats)
        self._bkgrms_stats = None  # delete to save memory
        result = self._apply_units(self._filter_grid(data))
        self._try_free_bkg_stats()
        return result

    @property
    @deprecated(since='3.0', alternative='n_pixels_mesh', until='4.0')
    def npixels_mesh(self):
        """
        A 2D array of the number pixels used to compute the statistics
        in each mesh.

        .. deprecated:: 3.0
            Use ``n_pixels_mesh`` instead.
        """
        return self._ngood

    @property
    def n_pixels_mesh(self):
        """
        A 2D array of the number pixels used to compute the statistics
        in each mesh.
        """
        return self._ngood

    @property
    @deprecated(since='3.0', alternative='n_pixels_map', until='4.0')
    def npixels_map(self):
        """
        A 2D map of the number of pixels used to compute the statistics
        in each mesh, resized to the shape of the input image.

        .. deprecated:: 3.0
            Use ``n_pixels_map`` instead.

        .. note::

            The returned image is (re)calculated each time this property
            is accessed. Store the result in a variable if you need to
            access it more than once.
        """
        return self.n_pixels_map

    @property
    def n_pixels_map(self):
        """
        A 2D map of the number of pixels used to compute the statistics
        in each mesh, resized to the shape of the input image.

        .. note::

            The returned image is (re)calculated each time this property
            is accessed. Store the result in a variable if you need to
            access it more than once.
        """
        n_pixels_map = block_replicate(self.n_pixels_mesh,
                                       self._interp_kwargs['box_size'],
                                       conserve_sum=False)
        return n_pixels_map[:self._interp_kwargs['shape'][0],
                            :self._interp_kwargs['shape'][1]]

    @lazyproperty
    def background_median(self):
        """
        The median value of the 2D low-resolution background map.

        This is equivalent to the value SourceExtractor prints to stdout
        (i.e., "(M+D) Background: <value>").

        .. note::

            This value is computed over the full ``background_mesh``,
            which includes IDW-interpolated values for any mesh boxes
            that were excluded from the statistics (e.g., due to masking
            or ``exclude_percentile``). It therefore represents the
            median of the final interpolated mesh, not solely the median
            of directly measured mesh values.
        """
        return self._apply_units(np.median(self.background_mesh))

    @lazyproperty
    def background_rms_median(self):
        """
        The median value of the low-resolution background RMS map.

        This is equivalent to the value SourceExtractor prints to stdout
        (i.e., "(M+D) RMS: <value>").

        .. note::

            This value is computed over the full
            ``background_rms_mesh``, which includes IDW-interpolated
            values for any mesh boxes that were excluded from the
            statistics (e.g., due to masking or ``exclude_percentile``).
            It therefore represents the median of the final interpolated
            mesh, not solely the median of directly measured mesh
            values.
        """
        return self._apply_units(np.median(self.background_rms_mesh))

    def _calculate_image(self, data):
        """
        Calculate the full-sized background or background rms image from
        the low-resolution mesh.
        """
        data = self.interpolator(data, **self._interp_kwargs)

        if self.coverage_mask is not None:
            data[self.coverage_mask] = self.fill_value

        return self._apply_units(data)

    @property
    def background(self):
        """
        A 2D `~numpy.ndarray` containing the background image.

        .. note::

            The returned image is (re)calculated each time this property
            is accessed. Store the result in a variable if you need to
            access it more than once.
        """
        return self._calculate_image(self.background_mesh)

    @property
    def background_rms(self):
        """
        A 2D `~numpy.ndarray` containing the background RMS image.

        .. note::

            The returned image is (re)calculated each time this property
            is accessed. Store the result in a variable if you need to
            access it more than once.
        """
        return self._calculate_image(self.background_rms_mesh)

    def plot_meshes(self, *, ax=None, marker='+', markersize=None,
                    color='blue', alpha=None, outlines=False, **kwargs):
        """
        Plot the low-resolution mesh boxes on a matplotlib Axes
        instance.

        Parameters
        ----------
        ax : `matplotlib.axes.Axes` or `None`, optional
            The matplotlib axes on which to plot. If `None`, then the
            current `~matplotlib.axes.Axes` instance is used.

        marker : str, optional
            The `matplotlib marker
            <https://matplotlib.org/stable/api/markers_api.html>`_ to
            use to mark the center of the boxes.

        markersize : float, optional
            The box center marker size in ``points ** 2``
            (typographical points are 1/72 inch) . The default is
            ``matplotlib.rcParams['lines.markersize'] ** 2``. If set to
            0, then the box center markers will not be plotted.

        color : str, optional
            The color for the box center markers and outlines.

        alpha : float, optional
            The alpha blending value, between 0 (transparent) and 1
            (opaque), for the box center markers and outlines.

        outlines : bool, optional
            Whether or not to plot the box outlines.

        **kwargs : dict, optional
            Any keyword arguments accepted by
            `matplotlib.patches.Patch`, which is used to draw the box
            outlines. Used only if ``outlines`` is True.
        """
        import matplotlib.pyplot as plt

        kwargs['color'] = color
        if ax is None:
            ax = plt.gca()

        mesh_xycen = np.flipud(self._calculate_mesh_yxcen())
        ax.scatter(*mesh_xycen, s=markersize, marker=marker, color=color,
                   alpha=alpha)

        if outlines:
            xycen = np.column_stack(mesh_xycen)
            apers = RectangularAperture(xycen, w=self.box_size[1],
                                        h=self.box_size[0], theta=0.0)
            apers.plot(ax=ax, alpha=alpha, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the _stats module.
"""

import importlib
import pickle  # nosec B403
from unittest.mock import patch

import astropy.units as u
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from photutils.utils._optional_deps import HAS_BOTTLENECK
from photutils.utils._stats import (nanmax, nanmean, nanmedian, nanmin, nanstd,
                                    nansum, nanvar)

funcs = [(nansum, np.nansum), (nanmean, np.nanmean),
         (nanmedian, np.nanmedian), (nanstd, np.nanstd),
         (nanvar, np.nanvar), (nanmin, np.nanmin), (nanmax, np.nanmax)]


@pytest.mark.skipif(not HAS_BOTTLENECK, reason='bottleneck is required')
@pytest.mark.parametrize('func', funcs)
@pytest.mark.parametrize('axis', [None, 0, 1, (0, 1), (1, 2), (2, 1),
                                  (0, 1, 2), (3, 1), (0, 3), (2, 0)])
@pytest.mark.parametrize('use_units', [False, True])
def test_nan_funcs(func, axis, use_units):
    """
    Test nan functions with various axes and unit combinations.
    """
    arr = np.ones((5, 3, 8, 9))
    if use_units:
        arr <<= u.m

    result1 = func[0](arr, axis=axis)
    result2 = func[1](arr, axis=axis)
    assert_equal(result1, result2)


@pytest.mark.skipif(not HAS_BOTTLENECK, reason='bottleneck is required')
@pytest.mark.parametrize('func', funcs)
def test_nan_funcs_float32(func):
    """
    Test that non-float64 arrays dispatch to numpy instead of
    bottleneck.
    """
    arr = np.ones((5, 3), dtype=np.float32)
    result1 = func[0](arr, axis=None)
    result2 = func[1](arr, axis=None)
    assert_equal(result1, result2)


@pytest.mark.parametrize('axis', [None, 0, 1])
@pytest.mark.parametrize('func', funcs)
def test_nan_function_pickle(axis, func):
    """
    Test that nan functions can be pickled and unpickled without error.

    This should work regardless of whether Bottleneck is installed.
    """
    nan_func = func[0]
    np_func = func[1]
    pickled_func = pickle.loads(  # noqa: S301
        pickle.dumps(nan_func))  # noqa: S301  # nosec B301
    data = np.arange(10).reshape(2, 5) ** 2
    result1 = nan_func(data, axis=axis)
    result2 = pickled_func(data, axis=axis)
    result3 = np_func(data, axis=axis)
    assert_allclose(result1, result2)
    assert_allclose(result2, result3)


def test_nanmean_repr():
    """
    Test that the repr of nanmean is correct.
    """
    result = repr(nanmean)
    assert 'nanmean' in result


def test_nan_funcs_no_bottleneck():
    """
    Test that the functions work when bottleneck is not available by
    reloading the module with HAS_BOTTLENECK mocked to False.
    """
    with patch('photutils.utils._optional_deps.HAS_BOTTLENECK',
               new=False):
        import photutils.utils._stats as stats_mod
        importlib.reload(stats_mod)

        arr = np.ones((5, 3))
        result = stats_mod.nansum(arr)
        assert_equal(result, np.nansum(arr))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Subpackage containing tools for measuring morphological properties of
objects in an astronomical image.
"""

from .core import *  # noqa: F401, F403
from .non_parametric import *  # noqa: F401, F403



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the _misc module.
"""

import builtins
from unittest.mock import patch

import pytest

from photutils.utils._misc import _get_date, _get_meta, _get_version_info


def test_get_version_info_import_error():
    """
    Test that _get_version_info returns None for packages that cannot
    be imported.
    """
    real_import = builtins.__import__

    def mock_import(name, *args, **kwargs):
        if name == 'gwcs':
            raise ImportError(name)
        return real_import(name, *args, **kwargs)

    with patch('builtins.__import__', side_effect=mock_import):
        versions = _get_version_info()
    assert versions['gwcs'] is None


@pytest.mark.parametrize('utc', [False, True])
def test_get_date_oserror_fallback(utc):
    """
    Test that _get_date falls back when astimezone raises OSError.
    """
    with patch('photutils.utils._misc.datetime') as mock_dt:
        mock_now = mock_dt.now.return_value
        mock_now.astimezone.side_effect = OSError('no timezone')
        mock_now.strftime.return_value = '2025-01-01 00:00:00'
        mock_dt.now.return_value = mock_now
        from datetime import UTC
        mock_dt.UTC = UTC
        result = _get_date(utc=utc)
        assert isinstance(result, str)


@pytest.mark.parametrize('utc', [False, True])
def test_get_meta(utc):
    """
    Test _get_meta returns expected keys.
    """
    meta = _get_meta(utc=utc)
    keys = ('date', 'version')
    for key in keys:
        assert key in meta

    versions = meta['version']
    assert isinstance(versions, dict)
    keys = ('Python', 'photutils', 'astropy', 'numpy', 'scipy', 'skimage',
            'matplotlib', 'gwcs', 'bottleneck')
    for key in keys:
        assert key in versions



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the cutouts module.
"""

import numpy as np
import pytest
from astropy.nddata import PartialOverlapError
from numpy.testing import assert_equal

from photutils.aperture import BoundingBox
from photutils.datasets import make_100gaussians_image
from photutils.utils.cutouts import CutoutImage, _make_cutouts


class TestCutoutImage:
    """
    Tests for the CutoutImage class.
    """

    def test_cutout(self):
        """
        Test CutoutImage with basic parameters.
        """
        data = make_100gaussians_image()
        shape = (24, 57)
        yxpos = (100, 51)
        cutout = CutoutImage(data, yxpos, shape)
        assert cutout.position == yxpos
        assert cutout.input_shape == shape
        assert cutout.mode == 'trim'
        assert np.isnan(cutout.fill_value)
        assert not cutout.copy

        assert cutout.data.shape == shape
        assert_equal(cutout.__array__(), cutout.data)

        assert isinstance(cutout.bbox_original, BoundingBox)
        assert isinstance(cutout.bbox_cutout, BoundingBox)
        assert cutout.slices_original == (slice(88, 112, None),
                                          slice(23, 80, None))
        assert cutout.slices_cutout == (slice(0, 24, None),
                                        slice(0, 57, None))

        assert_equal(cutout.xyorigin, np.array((23, 88)))

        cutouts2 = CutoutImage(data, yxpos, np.array(shape))
        assert cutouts2.input_shape == shape

        assert 'CutoutImage(' in repr(cutout)
        assert f'shape={shape}' in repr(cutout)
        assert f'Shape: {shape}' in str(cutout)

    def test_cutout_partial_overlap(self):
        """
        Test CutoutImage with partial overlap modes.
        """
        data = make_100gaussians_image()
        shape = (24, 57)

        # 'trim' mode
        cutout = CutoutImage(data, (11, 10), shape)
        assert cutout.input_shape == shape
        assert cutout.shape == (23, 39)

        # 'strict' mode
        match = 'Arrays overlap only partially'
        with pytest.raises(PartialOverlapError, match=match):
            CutoutImage(data, (11, 10), shape, mode='strict')

        # 'partial' mode
        cutout = CutoutImage(data, (11, 10), shape, mode='partial')
        assert cutout.input_shape == shape
        assert cutout.shape == shape

        assert (cutout.bbox_original
                == BoundingBox(ixmin=0, ixmax=39, iymin=0, iymax=23))
        assert (cutout.bbox_cutout
                == BoundingBox(ixmin=18, ixmax=57, iymin=1, iymax=24))

        assert cutout.slices_original == (slice(0, 23, None),
                                          slice(0, 39, None))
        assert cutout.slices_cutout == (slice(1, 24, None),
                                        slice(18, 57, None))

        assert_equal(cutout.xyorigin, np.array((-18, -1)))

        # Regression test for xyorgin in partial mode when cutout extends
        # beyond right or top edge
        data = make_100gaussians_image()
        shape = (54, 57)
        cutout = CutoutImage(data, (281, 485), shape, mode='partial')

        assert_equal(cutout.xyorigin, np.array((457, 254)))
        assert (cutout.bbox_original
                == BoundingBox(ixmin=457, ixmax=500, iymin=254, iymax=300))
        assert (cutout.bbox_cutout
                == BoundingBox(ixmin=0, ixmax=43, iymin=0, iymax=46))
        assert cutout.slices_original == (slice(254, 300, None),
                                          slice(457, 500, None))
        assert cutout.slices_cutout == (slice(0, 46, None),
                                        slice(0, 43, None))

    def test_cutout_copy(self):
        """
        Test CutoutImage with copy=True and copy=False.
        """
        data = make_100gaussians_image()

        cutout1 = CutoutImage(data, (1, 1), (3, 3), copy=True)
        cutout1.data[0, 0] = np.nan
        assert not np.isnan(data[0, 0])

        cutout2 = CutoutImage(data, (1, 1), (3, 3), copy=False)
        cutout2.data[0, 0] = np.nan
        assert np.isnan(data[0, 0])


class TestMakeCutouts:
    """
    Tests for the _make_cutouts utility function.
    """

    def setup_method(self):
        self.data = np.arange(100, dtype=float).reshape(10, 10)

    def test_fully_inside(self):
        """
        Test a source fully inside the image.
        """
        xpos = np.array([5.0])
        ypos = np.array([5.0])
        cutouts, mask = _make_cutouts(self.data, xpos, ypos, (3, 3))
        assert cutouts.shape == (1, 3, 3)
        np.testing.assert_array_equal(cutouts[0], self.data[4:7, 4:7])
        assert mask[0].all()

    def test_partial_overlap_corners(self):
        """
        Test sources at image corners that partially overlap.
        """
        xpos = np.array([0.0, 9.0])
        ypos = np.array([0.0, 9.0])
        _, mask = _make_cutouts(self.data, xpos, ypos, (5, 5))

        # Corner (0, 0): top-left 2 rows and 2 cols are outside
        assert not mask[0].all()  # not fully inside
        assert mask[0].any()  # not fully outside
        assert not mask[0, 0, 0]  # outside pixel
        assert mask[0, 2, 2]  # center pixel (the position itself)

        # Corner (9, 9): bottom-right 2 rows and 2 cols are outside
        assert not mask[1].all()
        assert mask[1].any()
        assert mask[1, 2, 2]  # center pixel
        assert not mask[1, 4, 4]  # outside pixel

    def test_no_overlap(self):
        """
        Test a source completely outside the image.
        """
        xpos = np.array([-10.0])
        ypos = np.array([-10.0])
        cutouts, mask = _make_cutouts(self.data, xpos, ypos, (3, 3))
        assert not mask[0].any()
        assert np.all(cutouts[0] == 0.0)

    def test_fill_value_nan(self):
        """
        Test that fill_value=NaN fills out-of-bounds pixels with NaN.
        """
        xpos = np.array([0.0])
        ypos = np.array([0.0])
        cutouts, mask = _make_cutouts(self.data, xpos, ypos, (5, 5),
                                      fill_value=np.nan)
        # Outside pixels should be NaN
        assert np.all(np.isnan(cutouts[0][~mask[0]]))
        # Inside pixels should not be NaN
        assert np.all(np.isfinite(cutouts[0][mask[0]]))

    def test_fill_value_custom(self):
        """
        Test that a custom fill_value is used for out-of-bounds pixels.
        """
        xpos = np.array([0.0])
        ypos = np.array([0.0])
        cutouts, mask = _make_cutouts(self.data, xpos, ypos, (3, 3),
                                      fill_value=-99.0)
        assert np.all(cutouts[0][~mask[0]] == -99.0)

    def test_overlap_mask_dtype(self):
        """
        Test that overlap_mask is a boolean array.
        """
        xpos = np.array([5.0])
        ypos = np.array([5.0])
        _, mask = _make_cutouts(self.data, xpos, ypos, (3, 3))
        assert mask.dtype == bool

    def test_mixed_sources(self):
        """
        Test a mix of fully-inside, partial, and outside sources.
        """
        xpos = np.array([5.0, 0.0, -10.0])
        ypos = np.array([5.0, 0.0, -10.0])
        _, mask = _make_cutouts(self.data, xpos, ypos, (3, 3))

        # Fully inside
        assert mask[0].all()
        # Partial overlap
        assert mask[1].any()
        assert not mask[1].all()
        # No overlap
        assert not mask[2].any()

    def test_even_shaped_cutout(self):
        """
        Test _make_cutouts with an even-shaped cutout.
        """
        xpos = np.array([5.0])
        ypos = np.array([5.0])
        cutouts, mask = _make_cutouts(self.data, xpos, ypos, (4, 4))
        assert cutouts.shape == (1, 4, 4)
        assert mask[0].all()  # fully inside
        # Half-widths: hy=2, hx=2; cutout rows [3..6], cols [3..6]
        expected = self.data[3:7, 3:7]
        np.testing.assert_array_equal(cutouts[0], expected)

    def test_even_shaped_cutout_at_edge(self):
        """
        Test _make_cutouts with an even-shaped cutout at the image edge.
        """
        xpos = np.array([0.0])
        ypos = np.array([0.0])
        cutouts, mask = _make_cutouts(self.data, xpos, ypos, (4, 4))
        assert cutouts.shape == (1, 4, 4)
        # Some pixels should be outside
        assert not mask[0].all()
        assert mask[0].any()
        # Outside pixels should be zero (default fill_value)
        assert np.all(cutouts[0][~mask[0]] == 0.0)

    def test_data_not_2d(self):
        """
        Test that a non-2D data array raises ValueError.
        """
        match = 'data must be a 2D array'
        with pytest.raises(ValueError, match=match):
            _make_cutouts(np.ones(10), np.array([5.0]),
                          np.array([5.0]), (3, 3))

    def test_xpos_not_1d(self):
        """
        Test that non-1D xpos/ypos arrays raise ValueError.
        """
        match = 'xpos and ypos must be 1D arrays'
        with pytest.raises(ValueError, match=match):
            _make_cutouts(self.data, np.ones((2, 2)),
                          np.array([5.0]), (3, 3))

    def test_cutout_shape_wrong_length(self):
        """
        Test that cutout_shape with != 2 elements raises ValueError.
        """
        match = 'cutout_shape must have exactly 2 elements'
        with pytest.raises(ValueError, match=match):
            _make_cutouts(self.data, np.array([5.0]),
                          np.array([5.0]), (3, 3, 3))

    def test_xpos_ypos_length_mismatch(self):
        """
        Test that mismatched xpos/ypos lengths raise ValueError.
        """
        match = 'xpos and ypos must have the same length'
        with pytest.raises(ValueError, match=match):
            _make_cutouts(self.data, np.array([5.0, 6.0]),
                          np.array([5.0]), (3, 3))
